#!/usr/bin/perl

## W3C W3C::Rdf::ListRdfParser - W3C PERL Library
# sketch of Eric's notion of an RDF schema to make it easier to handle
# combiniations of lists in RDF.
# 
# schema declaration and use:
# <?xml:namespace href="http://www.w3.org/schemas/listRDF" as="lst"?>
# <lst:region lst:confoundSet="a,b" lst:separators="[ \t\r\n,]+">
#
# four ways to describe the set of ("a1.b1", "a1.b2"):
#  <a>a1
#   <b>b1, b2</b>
#  </a>
#
#  and
#
#  <a>a1
#   <b>b1</b>
#   <b>b2</b>
#  </a>
#
#   and
#
#  <rdf:desc a="a1" b="b1">
#  <rdf:desc a="a1" b="b2">
#
#   and
#
#  <b>b1
#   <a>a1</a>
#  </b>
#  <b>b2
#   <a>a1</a>
#  </b>
#
# ending the region clears and suspends the recombination rules:
# </lst:region>
#
# A nested region lexically hides all the preceding tags. In this implementation,
# this is accomplished by stopping searching up the tree when a <lst:region> is
# found.
#
# Optimization: It is not my goal to write a version that is quick, so much as
# one that establishes useful shared protocols. Later, I will re-visit the code
# and put in more state to keep the engine from having to walk up the parent tree
# each time.

#Copyright Massachusetts Institute of technology, 1998.
#Written by Eric Prud'hommeaux

#things that need to be done:
#  write the damned thing

sub checkConfoundedSet {
    my $self = shift;
    my $startAt = shift;
    my @needed = @{$self->{CONFOUNDSET}};
    my %rdfElements = ();
    foreach my $needed (@{$self->{CONFOUNDSET}}) {
	my ($rdfElement, $type) = $self->{CURRENT}->findTagOrAttributeInside($needed, $self->{LISTRDFAS}.':region');
	return if (!defined $rdfElement || !defined $type);
	$ {$rdfElements{$needed}}{$rdfElement} = $rdfElement;
    }
    # if we reach here, we have one or more working combos.
    print "combo:\n";
    foreach my $needed (keys %rdfElements) {
	my $byElement = $rdfElements{$needed};
	foreach my $rdfElementName (keys %$byElement) {
	    my $rdfElement = $ {$byElement}{$rdfElementName};
	    my $type = $rdfElement->getType();
	    if ($type == $W3C::Rdf::RdfParser::RDFDATAELEMENT) {
		foreach my $dataChild (@{$rdfElement->{CHILDREN}}) {
		    my $element = $dataChild->getData() if ($dataChild->getName() eq '[DATA]');
		    foreach my $t (split($self->{SEPARATORS}, $element)) {
			$self->addCombo($needed, $rdfElement, $t);
		    }
		}
	    } elsif ($type == $W3C::Rdf::RdfParser::RDFELEMENT) {
		my $value = $rdfElement->{ATTRIBUTES}->getValue($needed);
		$self->addCombo($needed, $rdfElement, $value);
	    }
	}
    }
}

sub addCombo {
    my $self = shift;
    my $needed = shift;
    my $rdfElement = shift;
    my $value = shift;
    print "      $needed: $value\n";
#    push (@{$self->{LISTSET}}, $self->{CURRENTLIST}) if (!defined $ {$self->{COMBOS}}{$self->{CURRENTLIST}});
    $ {$ {$ {$self->{COMBOS}}{$self->{CURRENTLIST}}}{$needed}}{$rdfElement} = [] if (!exists $ {$self->{COMBOS}}{$needed});
    push (@{$ {$ {$ {$self->{COMBOS}}{$self->{CURRENTLIST}}}{$needed}}{$rdfElement}}, [$rdfElement, $value]);
#    $ {$self->{COMBOS}}{$needed} = [] if (!exists $ {$self->{COMBOS}}{$needed});
#    push (@{$ {$self->{COMBOS}}{$needed}}, $value);
}

sub getCombos {
    my $self = shift;
    return $self->{COMBOS};
}

1;
