#Copyright Massachusetts Institute of technology, 1998.
#Written by Eric Prud'hommeaux for the World Wide Web Consortium

use strict;
require Exporter;
#require AutoLoader;

$W3C::Rdf::RdfDB::REVISION = '$Id: RdfDB.pm,v 1.32 1999/12/14 13:14:40 eric Exp $ ';

package W3C::Rdf::RdfDB;
use vars qw($VERSION $DSLI @ISA @EXPORT @EXPORT_OK);
@ISA = qw(Exporter); # AutoLoader);
#@EXPORT = qw();
@EXPORT_OK = qw();
$VERSION = 0.94;
$DSLI = 'adpO';

use vars qw($VERSION $DSLI @ISA @EXPORT @EXPORT_OK
	    $NodeIdType_Fake $NodeIdType_Gen $NodeIdType_Ref $NodeIdType_String
	    $FixedId_Fake $FixedID_Gen $FixedID_Ref $FixedID_String
	    $ContainerIdType_Seq $ContainerIdType_Bag $ContainerIdType_Alt
	    $AttributionType_SOURCE $AttributionType_GEN $AttributionType_REI);
@EXPORT = qw($NodeIdType_Fake $NodeIdType_Gen $NodeIdType_Ref $NodeIdType_String
	     $ContainerIdType_Seq $ContainerIdType_Bag $ContainerIdType_Alt
	     $AttributionType_SOURCE $AttributionType_GEN $AttributionType_REI);
#($NodeIdType_ID, $NodeIdType_Ref, $NodeIdType_Gen, $NodeIdType_Fake, $NodeIdType_String) = 
#    ('ID', 'Ref', 'Gen', 'Fake', 'String');
($NodeIdType_Fake, $NodeIdType_Gen, $NodeIdType_Ref, $NodeIdType_String) = 
    ('Fake', 'Gen', 'Ref', 'String');
($FixedId_Fake, $FixedID_Gen, $FixedID_Ref, $FixedID_String) = 
    (9, 12, 11, 10);
($ContainerIdType_Seq, $ContainerIdType_Bag, $ContainerIdType_Alt) = 
    ('Seq', 'Bag', 'Alt');
($AttributionType_SOURCE, $AttributionType_GEN, $AttributionType_REI) = 
    ('Source', 'Generated', 'Reified');

package W3C::Rdf::Uri;

sub new {
    my ($proto, $uri) = @_;
    my $class = ref($proto) || $proto;
    my $self = {};
    bless ($self, $class);
    $self->{URI} = $uri;
    return $self;
}

sub show {
    my ($self, $prefix, $namespaceHandler) = @_;
    my $ret = $self->{URI};
    $ret = $namespaceHandler->unmapNamespace($ret) if ($namespaceHandler);
    return $ret;
}

package W3C::Rdf::String;

sub new {
    my ($proto, $string) = @_;
    my $class = ref($proto) || $proto;
    my $self = {};
    bless ($self, $class);
    $self->{STRING} = $string;
    return $self;
}

sub show {
    my ($self, $prefix, $namespaceHandler, $flags) = @_;
    my $ret = $self->{STRING};
    if (defined $flags->{-stringTrunc} && length $ret > $flags->{-stringTrunc}) {
	$ret = substr($ret, 0, $flags->{-stringTrunc}).$flags->{-truncFlag};
    }
    if ((my $wrap = $flags->{-stringWrap})) {
	my @ret;
	while ($ret =~ m/\G (.{$wrap}) /gcx) {
	    push (@ret, $1);
	}
	push (@ret, $1) if ($ret =~ m/\G (.*) $/gcx);
	$ret = join ("\n", @ret);
    }
    return $ret;
}

package W3C::Rdf::Attribution;

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $self = {};
    bless ($self, $class);
    ($self->{TRUSTED}, $self->{SOURCE}, $self->{PARENT}) = @_;
    eval {$self->{SOURCE}->show};
    if ($@) {
	print %$self;
    }
    return $self;
}

sub trusted {
    my ($self) = @_;
    return $self->{TRUSTED};
}

# mergeTrust - the rules for merging trust vary among the types of Attributions

sub mergeTrust {
    my ($self, $mergeWith) = @_;
    # trust only if this W3C::Rdf::Attribution and the one you are merging with are both trusted
    return $self->trusted & $mergeWith;
}

sub getSource {
    my ($self) = @_;
    return $self->{SOURCE};
}

sub getUri {
    my ($self) = @_;
    return $self->{SOURCE}->show;
}

# should probably come from the source rather than the type later on...
sub supplied {
    return 1;
}

sub trustedChain {
    my ($self, $namespaceHandler) = @_;
    my $source = $self->{SOURCE};
    if (ref $source) { # some are ATTRIBUTION_REIFICATION and ATTRIBUTION_PROPERTY
	$source = $source->show(undef, $namespaceHandler);
    }
    return ($self->trusted, $source) if (!$self->{PARENT});
    my ($pTrust, $pSource) = $self->{PARENT}->trustedChain($namespaceHandler);
    return ($self->mergeTrust($pTrust), $pSource.'->'.$source);
}

sub statedSum {
    my ($self, $statedSum) = @_;
    $self->{STATED_SUM} = $statedSum;
}

sub calculatedSum {
    my ($self, $calculatedSum) = @_;
    $self->{BAD_SUM} = $calculatedSum if (defined $self->{STATED_SUM} && $calculatedSum != $self->{STATED_SUM});
    if (defined $self->{STATED_SUM} && $calculatedSum != $self->{STATED_SUM}) {
#	print $self->show;
    }
}

sub show1 {
    my ($self, $marker, $prefix, $namespaceHandler) = @_;
    my ($trusted, $source) = ($self->trustedChain($namespaceHandler));
    my $sumStr = '';
    if (defined $self->{BAD_SUM}) {
	$sumStr = ' '.$self->{BAD_SUM}.' vs '.$self->{STATED_SUM}.' ';
    }
    return $prefix.$marker.$sumStr.($trusted ? ' ' : '#').'\''.$source.'\'';
}

sub dump {
    my ($self) = @_;
    return $self.':('.$self->{TRUSTED}.', '.$self->{SOURCE}.')';
}

package W3C::Rdf::SourceAttribution;
@W3C::Rdf::SourceAttribution::ISA = ('W3C::Rdf::Attribution');

sub show {
    my $self = shift;
    return $self->show1('s', @_);
}

sub _setSum ($) { # @@@ - needs a real API
}

sub flush { # @@@ - needs a real API
}

# W3C::Rdf::GeneratedAttribution - triples of predicate ::= rdf:type | rdf:_<n>

package W3C::Rdf::GeneratedAttribution;
@W3C::Rdf::GeneratedAttribution::ISA = ('W3C::Rdf::Attribution');

sub new {
    my ($proto, $trusted, $source, $parent) = @_;
    my $class = ref($proto) || $proto;
    $trusted = 1 if (!defined $trusted);
    $source = $parent->getSource if (!defined $source);
    my $self = $class->SUPER::new($trusted, $source, $parent);
    bless ($self, $class);
    return $self;
}

sub supplied {
    return 0;
}

sub show {
    my $self = shift;
    return $self->show1('g', @_);
}

package W3C::Rdf::ReifiedAttribution;
@W3C::Rdf::ReifiedAttribution::ISA = ('W3C::Rdf::Attribution');

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $reifiedTriple = shift;
    my ($trusted, $source) = (shift, shift);
    $trusted = 1 if (!defined $trusted);
    $source = $W3C::Rdf::RdfParser::ATTRIBUTION_REIFICATION if (!defined $source);
    my $self = $class->SUPER::new($trusted, $source, @_);
    $self->{REIFIED_TRIPLE} = $reifiedTriple;
    bless ($self, $class);
    return $self;
}

# reifications don't care about how their trust is derived - the statement was made.
sub mergeTrust {
    my ($self, $mergeWith) = @_;
    return $self->trusted;
}

sub show {
    my $self = shift;
    return $self->show1('r', @_);
}

package W3C::Rdf::BaseTriple;

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $self = {};
    bless ($self, $class);
    # this parameter stuffing is a convenience - don't count on these parms from subclass constructors
    ($self->{CONTAINER}, $self->{PREDICATE}, $self->{SUBJECT}, $self->{OBJECT}, $self->{ATTRIBUTION}) = @_;
#    print $self->dump."\n";
    return $self;
}

sub delete {
    my ($self) = @_;
    return $self->{CONTAINER}->deleteTriple($self);
}

sub setReificationInfo {
    my ($self, $nodeId, $container) = @_;
    $self->{REIFICATION_NODE} = $nodeId;
    $self->{REIFICATION_CONTAINER} = $container;
}

sub getReificationInfo {
    my ($self) = @_;
    return ($self->{REIFICATION_NODE}, $self->{REIFICATION_CONTAINER});
}

sub container {
    my ($self) = @_;
    return $self->{CONTAINER};
}

sub getPredicate {
    my ($self) = @_;
    return $self->{PREDICATE};
}

sub getSubject {
    my ($self) = @_;
    return $self->{SUBJECT};
}

sub getObject {
    my ($self) = @_;
    return $self->{OBJECT};
}

sub getAttribution {
    my ($self) = @_;
    return $self->{ATTRIBUTION};
}

sub trusted {
    my ($self) = @_;
    return $self->{ATTRIBUTION}->trusted;
}

sub dump {
    my ($self) = @_;
    return $self.':('.$self->{PREDICATE}.', '.$self->{SUBJECT}.', '.$self->{OBJECT}.') container:'.$self->{CONTAINER}.' attribution: '.$self->{ATTRIBUTION}->dump;
}

sub show {
    my ($self, $prefix, $namespaceHandler) = @_; $prefix = '  ' if (!defined $prefix);
    my ($predicate, $subject, $object) = ($self->getPredicate, $self->getSubject, $self->getObject);
    my ($reificationContainer, $reificationNode) = ($self->{REIFICATION_CONTAINER}, $self->{REIFICATION_NODE});
    $predicate = $predicate->show(undef, $namespaceHandler) if (ref $predicate);
    $subject = $subject->show(undef, $namespaceHandler) if (ref $subject);
    $object = $object->show(undef, $namespaceHandler) if (ref $object);

    my $ret;
    $ret .= $self->getAttribution->show($prefix, $namespaceHandler) if ($self->getAttribution);
    $ret .= ' triple(\''.$predicate.'\',\''.$subject.'\',\''.$object.'\')';
    if ($reificationContainer) {
	$reificationContainer = $reificationContainer->show(undef, $namespaceHandler);
	$reificationNode = $reificationNode->show(undef, $namespaceHandler);
	$ret .= ' reified in '.$reificationContainer.' as '.$reificationNode;
    }
    return $ret;
}

# triples that need the namespace mapped on the predicate (ie rdf -> http://www.w3.org/TR/rdf-syntax...)
package W3C::Rdf::MappedTriple;
@W3C::Rdf::MappedTriple::ISA = ('W3C::Rdf::BaseTriple');

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new(@_);
    bless ($self, $class);
    $self->{PREDICATE} = $self->{CONTAINER}->mapNamespace($self->{PREDICATE});
    return $self;
}

package W3C::Rdf::OrdinalTriple;
@W3C::Rdf::OrdinalTriple::ISA = ('W3C::Rdf::BaseTriple');

# W3C::Rdf::RdfConstantTriple - triples with a constant predicate

package W3C::Rdf::RdfConstantTriple;
@W3C::Rdf::RdfConstantTriple::ISA = ('W3C::Rdf::BaseTriple');

# I've experimented with a bunch of ways to implement what would be:
#   class Parent {
#       static theConst = 1;
#       public Client (Informer informer) {
#          return theConst;
#       }
#   };
#   class Child extends Parent {
#       static theConst = 2;
#   };


sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    # eval "\$${class}::Predicate" | $class == W3C::Rdf::SubjectTriple = $W3C::Rdf::SubjectTriple = ':subject'
    # this is a cheesy way to get at the class constant $class::Predicate. any better ideas?
#    my $self = $class->SUPER::new($_[0], $_[0]->{RDF_SCHEMA_URI}.eval "\$${class}::Predicate", @_[1..$#_]);
    my $self = $class->SUPER::new($_[0], '- replaced constant predicate -', @_[1..$#_]);
    $self->{PREDICATE} = $self->predicateConst;
    bless ($self, $class);
    return $self;
}

package W3C::Rdf::SubjectTriple;
@W3C::Rdf::SubjectTriple::ISA = ('W3C::Rdf::RdfConstantTriple');
#$W3C::Rdf::SubjectTriple::Predicate = $W3C::Rdf::RdfParser::PREDICATE_SUBJECT;
sub predicateConst {my ($self) = @_; return $self->{CONTAINER}->rdfToken($W3C::Rdf::RdfParser::PREDICATE_SUBJECT);}

package W3C::Rdf::PredicateTriple;
@W3C::Rdf::PredicateTriple::ISA = ('W3C::Rdf::RdfConstantTriple');
# $W3C::Rdf::PredicateTriple::Predicate = $W3C::Rdf::RdfParser::PREDICATE_PREDICATE;
sub predicateConst {my ($self) = @_; return $self->{CONTAINER}->rdfToken($W3C::Rdf::RdfParser::PREDICATE_PREDICATE);}

package W3C::Rdf::ObjectTriple;
@W3C::Rdf::ObjectTriple::ISA = ('W3C::Rdf::RdfConstantTriple');
# $W3C::Rdf::ObjectTriple::Predicate = $W3C::Rdf::RdfParser::PREDICATE_OBJECT;
sub predicateConst {my ($self) = @_; return $self->{CONTAINER}->rdfToken($W3C::Rdf::RdfParser::PREDICATE_OBJECT);}

package W3C::Rdf::TypeTriple;
@W3C::Rdf::TypeTriple::ISA = ('W3C::Rdf::RdfConstantTriple');
# $W3C::Rdf::TypeTriple::Predicate = $W3C::Rdf::RdfParser::PREDICATE_TYPE;

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
#    my $self = $class->SUPER::new(@_[0..1], $_[0]->{RDF_SCHEMA_URI}.eval "\$${class}::Object", @_[3..$#_]);
    my $self = $class->SUPER::new(@_[0..1], '- replaced constant object -', @_[2..$#_]);
    bless ($self, $class);
    return $self;
}

sub predicateConst {my ($self) = @_; return $self->{CONTAINER}->rdfToken($W3C::Rdf::RdfParser::PREDICATE_TYPE);}

package W3C::Rdf::MappedTypeTriple;
@W3C::Rdf::MappedTypeTriple::ISA = ('W3C::Rdf::TypeTriple');

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
#    my $self = $class->SUPER::new(@_[0..1], $_[0]->{RDF_SCHEMA_URI}.eval "\$${class}::Object", @_[3..$#_]);
    my $self = $class->SUPER::new(@_[0..1], @_[3..$#_]);
    $self->{OBJECT} = $self->{CONTAINER}->mapNamespace($_[2]);
    bless ($self, $class);
    return $self;
}

package W3C::Rdf::TypeStatementTriple;
@W3C::Rdf::TypeStatementTriple::ISA = ('W3C::Rdf::TypeTriple');
# $W3C::Rdf::TypeStatementTriple::Predicate = $W3C::Rdf::RdfParser::PREDICATE_TYPE;
# sub getPredicate {my ($self) = @_; return $self->{CONTAINER}->rdfToken($W3C::Rdf::RdfParser::PREDICATE_TYPE);}
# $W3C::Rdf::TypeStatementTriple::Object = $W3C::Rdf::RdfParser::OBJECT_STATEMENT;
sub getObject {my ($self) = @_; return $self->{CONTAINER}->rdfToken($W3C::Rdf::RdfParser::OBJECT_STATEMENT);}

package W3C::Rdf::TypeBagTriple;
@W3C::Rdf::TypeBagTriple::ISA = ('W3C::Rdf::TypeTriple');
# $W3C::Rdf::TypeBagTriple::Predicate = $W3C::Rdf::RdfParser::PREDICATE_TYPE;
# sub getPredicate {my ($self) = @_; return $self->{CONTAINER}->rdfToken($W3C::Rdf::RdfParser::PREDICATE_TYPE);}
# $W3C::Rdf::TypeBagTriple::Object = $W3C::Rdf::RdfParser::OBJECT_BAG;
sub getObject {my ($self) = @_; return $self->{CONTAINER}->rdfToken($W3C::Rdf::RdfParser::OBJECT_BAG);}

package W3C::Rdf::TypeSeqTriple;
@W3C::Rdf::TypeSeqTriple::ISA = ('W3C::Rdf::TypeTriple');
# $W3C::Rdf::TypeSeqTriple::Predicate = $W3C::Rdf::RdfParser::PREDICATE_TYPE;
# sub getPredicate {my ($self) = @_; return $self->{CONTAINER}->rdfToken($W3C::Rdf::RdfParser::PREDICATE_TYPE);}
# $W3C::Rdf::TypeSeqTriple::Object = $W3C::Rdf::RdfParser::OBJECT_SEQ;
sub getObject {my ($self) = @_; return $self->{CONTAINER}->rdfToken($W3C::Rdf::RdfParser::OBJECT_SEQ);}

package W3C::Rdf::TypeAltTriple;
@W3C::Rdf::TypeAltTriple::ISA = ('W3C::Rdf::TypeTriple');
# $W3C::Rdf::TypeAltTriple::Predicate = $W3C::Rdf::RdfParser::PREDICATE_TYPE;
# sub getPredicate {my ($self) = @_; return $self->{CONTAINER}->rdfToken($W3C::Rdf::RdfParser::PREDICATE_TYPE);}
# $W3C::Rdf::TypeAltTriple::Object = $W3C::Rdf::RdfParser::OBJECT_ALT;
sub getObject {my ($self) = @_; return $self->{CONTAINER}->rdfToken($W3C::Rdf::RdfParser::OBJECT_ALT);}

package W3C::Rdf::RdfObject;

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $self = {};
    bless ($self, $class);
    ($self->{TYPE_STR}, $self->{RDF_DB}) = @_;
    return $self;
}

sub mapNamespace {
    my ($self) = (shift);
    return $self->{RDF_DB}->mapNamespace(@_);
}

sub show {
    my ($self, $prefix) = @_; $prefix = '  ' if (!defined $prefix);
    my $ret = '';
    $ret .= $prefix.$self."\n"; # 'Type: '.$self->{TYPE_STR}."\n";
    $ret .= $prefix.'RDF DB: '.$self->{RDF_DB}."\n";
    return $ret;
}

package W3C::Rdf::RdfContainer;
@W3C::Rdf::RdfContainer::ISA = ('W3C::Rdf::RdfObject');

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my ($type, $rdfDB, $containerId) = @_;
    my $self = $class->SUPER::new($type, $rdfDB);
    $self->{ID} = $containerId;
    $self->{NEXT_ORDINAL} = 1;
    $self->{SUBJECTS} = [];
    return $self;
}

sub showNUKE {
    my ($self, $prefix) = @_; $prefix = '  ' if (!defined $prefix);
    my $ret = $self->SUPER::show($prefix);
    $ret .= $prefix.'ID: '.$self->{ID}->show."\n";
    $ret .= $prefix.'NEXT_ORDINAL: '.$self->{NEXT_ORDINAL}."\n";
    $ret .= $prefix.'SUBJECTS: '.join(', ', @{$self->{SUBJECTS}})."\n";
    return $ret;
}

package W3C::Rdf::RdfSeq;
@W3C::Rdf::RdfSeq::ISA = ('W3C::Rdf::RdfContainer');

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new('Seq', @_);
    bless ($self, $class);
    return $self;
}

package W3C::Rdf::RdfBag;
@W3C::Rdf::RdfBag::ISA = ('W3C::Rdf::RdfContainer');

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new('Bag', @_);
    bless ($self, $class);
    return $self;
}

package W3C::Rdf::RdfAlt;
@W3C::Rdf::RdfAlt::ISA = ('W3C::Rdf::RdfContainer');

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new('Alt', @_);
    bless ($self, $class);
    return $self;
}

package W3C::Rdf::RdfDescription;
@W3C::Rdf::RdfDescription::ISA = ('W3C::Rdf::RdfBag');

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my ($rdfDB, $containerId, $type, $about, $aboutEach, $aboutEachPrefix) = @_;
    my $self = $class->SUPER::new($rdfDB, $containerId);
    bless ($self, $class);
    ($self->{TYPE}, $self->{ABOUT}, $self->{ABOUT_EACH}, $self->{ABOUT_EACH_PREFIX}) = 
	($type, $about, $aboutEach, $aboutEachPrefix);
    return $self;
}

sub show {
    my ($self, $prefix) = @_; $prefix = '  ' if (!defined $prefix);
    my $ret = $self->SUPER::show($prefix);
    $ret .= $prefix.'ABOUT_CONTAINERS: '.join(', ', @{$self->{ABOUT_CONTAINERS}})."\n";
    $ret .= $prefix.'PREDICATE_OBJECTS: ';
    foreach my $po (@{$self->{PREDICATE_OBJECTS}}) {
	$ret .= join (' => ', @$po);
    }
    $ret .= "\n";
    return $ret;
}

package W3C::Rdf::RdfDB::MappedNode;
sub new ($$$$$) {
    my ($proto, $aboutEach, $mapFrom, $description, $bagId, $type, $attribution) = @_;
    my $class = ref($proto) || $proto;
    my $self = {};
    bless ($self, $class);
    ($self->{ABOUTEACH}, $self->{MAPFROM}, $self->{DESCRIPTION}, $self->{BAGID}, $self->{TYPE}, $self->{ATTRIBUTION}) = 
	($aboutEach, $mapFrom, $description, $bagId, $type, $attribution);
    return $self;
}

sub mapToTriple ($) {
    my ($self, $triple) = @_;
    if ($triple->getPredicate == $self->getPredicate && $triple->getSubject == $self->getSubject) {
	return $triple->getObject;
    } else {
	return $self;
    }
}

sub show ($;$$\%) {
    my ($self, $prefix, $namespaceHandler, $flags) = @_; $prefix = '  ' if (!defined $prefix); $flags = {} if (!defined $flags);

    my $ret;
    $ret .= '(\''.$self->{MAPFROM}->show(undef, $namespaceHandler);
    $ret .= '\',\''.$self->{ABOUTEACH}->show(undef, $namespaceHandler);
    $ret .= '\',*)';
#    $ret .= ' for '.$self->{DESCRIPTION}->{ID}->show($namespaceHandler) if (defined $self->{DESCRIPTION});
#    $ret .= $self->{ATTRIBUTION}->show($namespaceHandler);
    return $ret;
}

sub getPredicate () {my ($self) = @_; return $self->{MAPFROM};}
sub getSubject () {my ($self) = @_; return $self->{ABOUTEACH};}
sub getObject () {my ($self) = @_; return $self->{BOGUS};} # @@@

package W3C::Rdf::BaseTriple;

sub getReification {
    my ($self) = @_;
    my ($reificationContainer, $reificationNode) = ($self->{REIFICATION_CONTAINER}, $self->{REIFICATION_NODE});
    return () if (!$reificationNode);
    my @ret;
    my $reificationAttribution = new W3C::Rdf::ReifiedAttribution($self, 1, undef, $self->getAttribution);
    push (@ret, new W3C::Rdf::PredicateTriple($self->{CONTAINER}, $reificationNode, $self->getPredicate, $reificationAttribution));
    push (@ret, new W3C::Rdf::SubjectTriple($self->{CONTAINER}, $reificationNode, $self->getSubject, $reificationAttribution));
    push (@ret, new W3C::Rdf::ObjectTriple($self->{CONTAINER}, $reificationNode, $self->getObject, $reificationAttribution));
    push (@ret, new W3C::Rdf::TypeStatementTriple($self->{CONTAINER}, $reificationNode, $reificationAttribution));
#    push (@ret, $reificationContainer->addOrdinalTriple($reificationNode, undef)) if ($reificationContainer);
    push (@ret, new W3C::Rdf::OrdinalTriple($reificationContainer, 
					    $self->{RDF_DB}->{RDF_SCHEMA_URI}.'_'.$reificationContainer->nextOrdinal, 
					    $reificationContainer->id, 
					    undef, 
					    $reificationAttribution));
#					    new W3C::Rdf::GeneratedAttribution(1, undef, 
#									       $self->{RDF_DB}->{CURRENT_SOURCE_ATTRIBUTION})));
    return @ret;
}

package W3C::Rdf::RdfDB::MappedStatement;
@W3C::Rdf::RdfDB::MappedStatement::ISA = ('W3C::Rdf::BaseTriple');

sub new {
    my ($proto, $predicate, $subject, $object, $container, $reification, $attribution) = @_;
    my $class = ref($proto) || $proto;
    my $self = {PREDICATE => $predicate, SUBJECT => $subject, OBJECT => $object, 
		CONTAINER => $container, REIFICATION => $reification, ATTRIBUTION => $attribution};
    bless ($self, $class);
    return $self;
}

sub mapToTriple ($) {
    my ($self, $triple) = @_;
    my $ret = new W3C::Rdf::RdfDB::MappedStatement($self->mapNodeToTriple($self->getPredicate, $triple), 
						   $self->mapNodeToTriple($self->getSubject, $triple), 
						   $self->mapNodeToTriple($self->getObject, $triple), 
						   $self->getContainer, 
						   $self->getReification, 
						   $self->getAttribution);
    return $ret;
}

sub mapNodeToTriple ($$) {
    my ($self, $node, $triple) = @_;
    if ($node->isa('W3C::Rdf::RdfDB::MappedNode')) {
	return $node->mapToTriple($triple);
    } else {
	return $node;
    }
}

sub getPredicate () {my ($self) = @_; return $self->{PREDICATE};}
sub getSubject () {my ($self) = @_; return $self->{SUBJECT};}
sub getObject () {my ($self) = @_; return $self->{OBJECT};}
sub getContainer () {my ($self) = @_; return $self->{CONTAINER};}
sub getReification () {my ($self) = @_; return $self->{REIFICATION};}
sub getAttribution () {my ($self) = @_; return $self->{ATTRIBUTION};}

# quick and dirty W3C::Rdf::RdfDBException
package W3C::Rdf::RdfDBException;
use W3C::SAX::SAXException;
@W3C::Rdf::RdfDBException::ISA = ('W3C::SAX::SAXException');

use W3C::Rdf::Algernon;
package W3C::Rdf::RdfDB;
@W3C::Rdf::RdfDB::ISA = qw(W3C::Rdf::Algernon);

#####
# per-object data

#####
# new - prepare an empty W3C::Rdf::RdfDB

sub new {
    my ($proto, $parms) = @_;
    my $class = ref($proto) || $proto;
    my $self = {};
    bless ($self, $class);

    $self->{ERROR_HANDLER} = $parms->{-errorHandler};
    $self->{LAZY_REIFICATION} = $parms->{-lazyReification};
    $self->{OBSERVERS} = [];
    $self->{RDF_SCHEMA_URI} = 'http://www.w3.org/1999/02/22-rdf-syntax-ns#';
    $self->clear;
    return $self;
}

sub clear {
    my ($self) = @_;
    $self->{TRIPLES} = [];
    $self->{BY_CONTENTS} = {};
    $self->{ABOUT_EACH} = {};
    $self->{ABOUT_EACH_PREFIX} = {};
    $self->{CONTAINERS} = {};
    $self->{NAMESPACES} = {};
    $self->{NEXT_ASSIGNED_ID} = 1;
    $self->{RDF_PREFIX} = 'rdf'; # default to slurping rdf '<'; # cheap way to guarantee no match until initialized
#    $self->{CURRENT_PROPERTY_ATTRIBUTION} = new W3C::Rdf::GeneratedAttribution(1, $W3C::Rdf::RdfParser::ATTRIBUTION_PROPERTY);
}

%W3C::Rdf::RdfDB::Uris = ();
sub getUri ($) {
    my ($self, $uri) = @_;
    my $ret = $W3C::Rdf::RdfDB::Uris{$uri};
    if (!$ret) {
	$ret = new W3C::Rdf::Uri($uri);
	$W3C::Rdf::RdfDB::Uris{$uri} = $ret;
    }
    return $ret;
}

%W3C::Rdf::RdfDB::Strings = ();
sub getString ($) {
    my ($self, $string) = @_;
    my $ret = $W3C::Rdf::RdfDB::Strings{$string};
    if (!$ret) {
	$ret = new W3C::Rdf::String($string);
	$W3C::Rdf::RdfDB::Strings{$string} = $ret;
    }
    return $ret;
}

sub createGenId {
    my ($self, $attribution) = @_;
    return $self->getUri($attribution->{SOURCE}->{URI}.'#genid'.$self->{NEXT_ASSIGNED_ID}++);
}

%W3C::Rdf::RdfDB::Attributions = ();
sub getAttribution ($) {
    my ($self, $type, $uri, $parent, $source, $trusted, $clear) = @_;
    my $ret = $W3C::Rdf::RdfDB::Attributions{$type}{$uri}{$parent}{$source}{$trusted}{$clear};
    if (!$ret) {
	if ($type eq $AttributionType_SOURCE) {
	    $ret = new W3C::Rdf::SourceAttribution($trusted, $uri, $parent);
	} elsif ($type eq $AttributionType_GEN) {
	    $ret = new W3C::Rdf::GeneratedAttribution($trusted, $uri, $parent);
	} elsif ($type eq $AttributionType_REI) {
	    $ret = new W3C::Rdf::ReifiedAttribution($trusted, $uri, $parent);
	} else {
	    die 'type:'.$type;
	}
	$W3C::Rdf::RdfDB::Attributions{$type}{$uri}{$parent}{$source}{$trusted}{$clear} = $ret;
    }
    return $ret;
}

sub getSourceAttribution {
    my ($self) = @_;
    return $self->{CURRENT_SOURCE_ATTRIBUTION};
}

sub setSourceAttribution {
    my ($self, $sourceAttribution) = @_;
    $self->{CURRENT_SOURCE_ATTRIBUTION} = $sourceAttribution;
}

sub setErrorHandler {
    my $self = shift;
    $self->{ERROR_HANDLER} = shift;
}

sub setSource {
    my ($self, $source) = @_;
    $self->{SOURCE} = $source;
}

use URI::URL;

sub absoluteLink {
    my ($self, $externalId) = (shift, shift);
    $externalId = $externalId->show if ($externalId);
    foreach my $ref (@_) {
	if (defined $$ref) {
	    my $uri = new URI::URL($$ref, $externalId)->abs->as_string;
	    $$ref = $self->getUri($uri);
	}
    }
}

sub absoluteName { # assumes fragment (automaticly prepend '#') - maybe should even enforce it
    my ($self, $externalId) = (shift, shift);
    $externalId = $externalId->show if ($externalId);
    foreach my $ref (@_) {
	if (defined $$ref) {
	    my $uri = new URI::URL('#'.$$ref, $externalId)->abs->as_string;
	    $$ref = $self->getUri($uri);
	}
    }
}

sub absoluteString {
    my ($self) = (shift);
    foreach my $ref (@_) {
	if (defined $$ref) {
	    my $string = $self->getString($$ref);
	    $$ref = $string;
	}
    }
}

sub setRdfPrefix {
    my ($self, $newUri) = @_;
    $self->{RDF_SCHEMA_URI} = $newUri;
}

sub getRdfPrefix {
    my ($self) = @_;
    return $self->{RDF_SCHEMA_URI}; # !!! $self->{RDF_PREFIX};
}

sub rdfToken ($) {
    my ($self, $fragment) = @_;
    return $self->getUri('http://www.w3.org/1999/02/22-rdf-syntax-ns'.'#'.$fragment); # $self->{RDF_SCHEMA_URI}); @@@
}

# namespace support - still in flux: not sure whether namespaces will
# be expanded by downcalls from the XML parser or by a W3C::SAX::HandlerStream.
# Those marked with a !!! change depending on which interface is used.

sub addNamespace {
    my ($self, $sAs, $sHref) = @_;
    my $lower = $sAs;
    $lower =~ tr/A-Z/a-z/;
    $self->{NAMESPACES}->{$lower} = $sHref;
    $self->{RDF_PREFIX} = $sAs if ($sHref eq $self->{RDF_SCHEMA_URI});
}

sub getNamespace {
    my ($self, $sAs) = @_;
    my $lower = $sAs;
    $lower =~ tr/A-Z/a-z/;
    return $self->{NAMESPACES}->{$lower};
}

sub mapNamespace {
    my ($self, $predicate) = @_;
    return $predicate; # !!!
    my ($namespace, $property) = $predicate =~ m/(?:([^:]*):)?(.*)/;
    return $self->{NAMESPACES}{$namespace}.$property;
}

sub createDescription {
    my ($self, $type, $about, $aboutEach, $aboutEachPrefix, $bagId, $attribution) = @_;
    my $description = new W3C::Rdf::RdfDescription($self, $bagId, $type, $about, $aboutEach, $aboutEachPrefix);
    $self->addTriple(new W3C::Rdf::TypeBagTriple($self, $bagId, new W3C::Rdf::GeneratedAttribution(1, undef, $attribution)));
    my $node;
    if ($about) {
	$node = $about;
	$self->addTriple(new W3C::Rdf::MappedTypeTriple($self, $about, $type, $self->{CURRENT_SOURCE_ATTRIBUTION})) if ($type);
    } elsif ($aboutEach) {
	$node = new W3C::Rdf::RdfDB::MappedNode($aboutEach, 
						$self->rdfToken($W3C::Rdf::RdfParser::PREDICATE_TYPE_LI), 
						$description, $bagId, $type, $attribution);
    } elsif ($aboutEachPrefix) {
	die;
    }
    return ($description, $node);
}

sub expandMappedStatements (\@$$$) {
    my ($self, $mappedStatements, $container, $reification, $attribution) = @_;
    foreach my $mappedStatement (@$mappedStatements) {
	my @predicates = $self->triplesMatchingMappedNode($mappedStatement->getPredicate);
	my @subjects = $self->triplesMatchingMappedNode($mappedStatement->getSubject);
	my @objects = $self->triplesMatchingMappedNode($mappedStatement->getObject);
	foreach my $predicate (@predicates) {
	    foreach my $subject (@subjects) {
		foreach my $object (@objects) {
		    $self->addStatement($predicate, $subject, $object, $container, $reification, $attribution);
		}
	    }
	}
    }
}

sub addStatement ($$$$$$) {
    my ($self, $predicate, $subject, $object, $container, $reification, $attribution) = @_;
    if (!ref $predicate || !ref $subject || !ref $object) {
	return;
    }
    if ($self->hasMappedNodes($predicate, $subject, $object)) {
	my $mappedStatement = $self->createMappedStatement($predicate, $subject, $object, 
							   $container, $reification, $attribution);
	$self->expandMappedStatements([$mappedStatement], $container, $reification, $attribution);
    } else {
	my ($all, $triple) = $self->createStatement($predicate, $subject, $object, $container, $reification, $attribution);
	my @matchingMappedStatements = $self->mappedStatementsMatchingTriple($triple);
	foreach my $mappedStatement (@matchingMappedStatements) {
	    $self->expandMappedStatements([$mappedStatement->mapToTriple($triple)], $container, $reification, $attribution);
	}
	return $triple;
    }
}

sub hasMappedNodes ($$$) {
    my ($self, $predicate, $subject, $object) = @_;
    my ($mappedPredicate, $mappedSubject, $mappedObject) = 
	($predicate->isa(qw(W3C::Rdf::RdfDB::MappedNode)), 
	 $subject->isa(qw(W3C::Rdf::RdfDB::MappedNode)), 
	 $object->isa(qw(W3C::Rdf::RdfDB::MappedNode)));
    return $mappedPredicate || $mappedSubject || $mappedObject;
}

sub mappedStatementsMatchingTriple ($) {
    my ($self, $triple) = @_;
    my @ret;
    foreach my $mappedStatement (@{$self->{MAPPED_STATEMENTS}}) {
	if (($mappedStatement->getPredicate->isa(qw(W3C::Rdf::RdfDB::MappedNode)) && 
	     $triple->getPredicate == $mappedStatement->getPredicate->getPredicate && 
	     $triple->getSubject == $mappedStatement->getPredicate->getSubject) || 
	    ($mappedStatement->getSubject->isa(qw(W3C::Rdf::RdfDB::MappedNode)) && 
	     $triple->getPredicate == $mappedStatement->getSubject->getPredicate && 
	     $triple->getSubject == $mappedStatement->getSubject->getSubject) || 
	    ($mappedStatement->getObject->isa(qw(W3C::Rdf::RdfDB::MappedNode)) && 
	     $triple->getPredicate == $mappedStatement->getObject->getPredicate && 
	     $triple->getSubject == $mappedStatement->getObject->getSubject)) {
	    push (@ret, $mappedStatement);
	}
    }
    return @ret;
}

sub triplesMatchingMappedNode ($) {
    my ($self, $potentialMappedNode) = @_;
    if ($potentialMappedNode->isa('W3C::Rdf::RdfDB::MappedNode')) {
	my @ret;
	foreach my $triple (@{$self->{TRIPLES}}) {
	    if ($triple->getPredicate == $potentialMappedNode->getPredicate && # @@@
		$triple->getSubject == $potentialMappedNode->getSubject) {
		push (@ret, $triple->getObject);
	    }
	}
	return @ret;
    } else {
	return ($potentialMappedNode);
    }
}

sub createStatement ($$$$$$) {
    my ($self, $predicate, $subject, $object, $container, $reification, $attribution) = @_;
    my $triple = new W3C::Rdf::MappedTriple($self, $predicate, $subject, $object, $attribution);
    my @all;
    push (@all, $self->addTriple($triple));
    push (@all, $self->reify($container, $triple, $reification));
    return (\@all, $triple);
}

sub createMappedStatement ($$$$$$) {
    my ($self, $predicate, $subject, $object, $container, $reification, $attribution) = @_;
    my $mappedStatement = new W3C::Rdf::RdfDB::MappedStatement($predicate, $subject, $object, 
							       $container, $reification, $attribution);
    push (@{$self->{MAPPED_STATEMENTS}}, $mappedStatement);
    return $mappedStatement;
}

#####
# public triple manipulators

sub addTriple {
    my ($self, $triple) = @_;

    # do some sanity checks, at least while we are in the debugging stage
    if (!ref $triple || 
	!$triple->isa('W3C::Rdf::BaseTriple') || 
	!defined $triple->getPredicate || !ref $triple->getPredicate || # $triple->getPredicate eq '' ||
	!defined $triple->getSubject || !ref $triple->getSubject || # $triple->getSubject eq '' ||
	!defined $triple->getObject || !ref $triple->getObject || # $triple->getObject eq '' ||
	!ref $triple->getAttribution || 
	!$triple->{ATTRIBUTION}->isa('W3C::Rdf::Attribution')) {
	$self->{ERROR_HANDLER}->fatalError(new W3C::Rdf::RdfDBException("added bogus triple $triple to triple database: $self"));
    }
    if (exists $self->{BY_CONTENTS}{$triple->getPredicate->show}{$triple->getSubject->show}{
	$triple->getObject}{$triple->getAttribution->show}) {
	$self->{ERROR_HANDLER}->warning(new W3C::Rdf::RdfDBException('added duplicage triple '.$triple->show.' to triple database: '.$self));
    }

    # notify interested observers
    foreach my $observerEntry (@{$self->{OBSERVERS}}) {
	my ($triples, $observer) = @$observerEntry;
	# leverage off the already written triplesMatching
	$observer->notifyTriple($triple) if $self->triplesMatching([$triple], $triples);
    }

    push (@{$self->{TRIPLES}}, $triple);
    $self->{BY_CONTENTS}{$triple->getPredicate->show}{$triple->getSubject->show}{
	$triple->getObject}{$triple->getAttribution->show} = $triple;
    return $triple;
}

# passed a single triple or a HASH of triples
# return: # - number of deleted elements

sub deleteTriple {
    my ($self, $triples) = @_;
    my $start = $#{$self->{TRIPLES}};
    my $doomed = {};
    if (ref $triples eq 'HASH') {
	$doomed = $triples;
    } elsif (ref $triples eq 'ARRAY') {
	foreach my $triple (@$triples) {
	    $doomed->{$triple} = $triple;
	}
    } elsif (ref $triples eq '') {
	$doomed->{$triples} = $triples;
    }
    $self->{TRIPLES} = [grep {
	if ($doomed->{$_}) {
	    delete $self->{BY_CONTENTS}{$_->getPredicate.$_->getSubject.$_->getObject};
	    0;
	} else {
	    1;
	}} @{$self->{TRIPLES}}];
    return $start - $#{$self->{TRIPLES}};
}

sub reify {
    my ($self, $container, $triple, $nodeId) = @_;
    $nodeId ||= $self->assignId;
    $triple->setReificationInfo($nodeId, $container);
    my @ret;
    @ret = $self->reallyReify($triple) if (!$self->{LAZY_REIFICATION});
    return @ret;
}

sub reallyReify {
    my ($self, $triple) = @_;
    my ($nodeId, $container) = $triple->getReificationInfo;
    return () if (!$nodeId);
    my $reificationAttribution = new W3C::Rdf::ReifiedAttribution($triple, 1, undef, $triple->getAttribution);
    my @ret;
#    if (exists $self->{REIFICATIONS}{$triple}) {
#	$self->{ERROR_HANDLER}->error(new W3C::Rdf::RdfDBException('re-reified triple '.$triple->show.' in '.$self));
#    }
#    $self->{REIFICATIONS}{$triple} = $triple;
    push (@ret, $self->addTriple(new W3C::Rdf::PredicateTriple($self, $nodeId, $triple->getPredicate, $reificationAttribution)));
    push (@ret, $self->addTriple(new W3C::Rdf::SubjectTriple($self, $nodeId, $triple->getSubject, $reificationAttribution)));
    push (@ret, $self->addTriple(new W3C::Rdf::ObjectTriple($self, $nodeId, $triple->getObject, $reificationAttribution)));
    push (@ret, $self->addTriple(new W3C::Rdf::TypeStatementTriple($self, $nodeId, $reificationAttribution)));
#    push (@ret, $container->addOrdinalTriple($nodeId, undef, $reificationAttribution)) if ($container); @@@
    return @ret;
}

sub setLazyReification {
    my ($self, $newState) = @_;
    return () if ($newState == $self->{LAZY_REIFICATION});
    $self->{LAZY_REIFICATION} = $newState;
    my @ret;
    if ($newState) {
	my %doomed;
	foreach my $triple (@{$self->{TRIPLES}}) {
	    $doomed{$triple} = $triple if ($triple->getAttribution->isa('W3C::Rdf::ReifiedAttribution'));
	}
	$self->deleteTriple(\%doomed);
    } else {
	foreach my $triple (@{$self->{TRIPLES}}) {
	    push (@ret, $self->reallyReify($triple));
	}
    }
    return @ret;
}

#####
# triplet database callbacks

# These slow the addTriple method down so they should only be used
# where needed because the parsing needs to change as a result of
# added triples, as in the case of the trust mechanism in W3C::Rdf::RdfParser.


sub registerTripleObserver {
    my ($self, $triples, $observer) = @_;
    push (@{$self->{OBSERVERS}}, [$triples, $observer]);
}

#####
# triplet database queries

#####
# supplied - strain view for supplied triples (ones that were added to DB rather than reified or something like that

sub supplied {
    my ($self, $view) = @_;
    $view ||= $self->{TRIPLES};
    my @ret;
    foreach my $triple (@$view) {
	push (@ret, $triple) if ($triple->getAttribution->supplied);
    }
    return @ret;
}

#####
# triplesMatching (lookFor) - returns all triplets that match any of the triples in @$lookFors
# ex:	my @resourceList = $rdfDB->triplesMatching(undef, [['http://www.w3.org/schema/certHTMLv1/hasAccessTo', $node, undef]]);
# bug - ignores attribution filter
sub triplesMatching {
    my ($self, $view, $lookFors, $flags) = @_;
    $view ||= $self->{TRIPLES};
    my @ret;
  TRIPLE:
    foreach my $triple (@$view) {
	foreach my $lookFor (@$lookFors) {
	    if ((!defined $lookFor->[0] || $triple->getPredicate eq $lookFor->[0]) &&
		(!defined $lookFor->[1] || $triple->getSubject eq $lookFor->[1]) &&
		(!defined $lookFor->[2] || $triple->getObject eq $lookFor->[2]) && 
		(!$flags->{-sourceOnly} || $triple->getAttribution->supplied) && 
		(!$flags->{-attributions} || scalar grep {$_ == $triple->getAttribution} @{$flags->{-attributions}})) {
		push (@ret, $triple);
		next TRIPLE; # don't bother checking the remaining lookFors, we've already added the triple
	    }
	}
    }
    return @ret;
}

#####
# anySubjects (lookFor) - returns all triplets with this subject lookFor

sub anySubjects {
    my ($self, $view, $lookFors) = @_;
    $view ||= $self->{TRIPLES};
    my @ret;
  TRIPLE:
    foreach my $triple (@$view) {
	foreach my $lookFor (@$lookFors) {
	    if ($triple->getSubject eq $lookFor) {
		push (@ret, $triple);
		next TRIPLE; # don't bother checking the remaining lookFors, we've already added the triple
	    }
	}
    }
    return @ret;
}

sub anyObjects {
    my ($self, $view, $lookFors) = @_;
    $view ||= $self->{TRIPLES};
    my @ret;
  TRIPLE:
    foreach my $triple (@$view) {
	foreach my $lookFor (@$lookFors) {
	    if ($triple->getObject eq $lookFor) {
		push (@ret, $triple);
		next TRIPLE; # don't bother checking the remaining lookFors, we've already added the triple
	    }
	}
    }
    return @ret;
}

# return the list of Containers that contain the elements in lookFors.
sub anyContainerNodes {
    my ($self, $view, $lookFors) = @_;
    $view ||= $self->{TRIPLES};
    my @ret;
    foreach my $rdfElement ($self->anyObjects($view, $lookFors)) {
	push (@ret, $rdfElement) if ($rdfElement->getPredicate->show =~ m/\A \Q$self->{RDF_SCHEMA_URI}\E (?:(li) | (?:_ (\d+))) \Z/x);
    }
    return @ret;
}

sub nodesWithType {
    my ($self, $view, $lookFors) = @_;
    $view ||= $self->{TRIPLES};
    $lookFors = [$lookFors] unless (ref $lookFors eq 'ARRAY');
    my @ret;
  TRIPLE:
    foreach my $triple (@$view) {
	foreach my $lookFor (@$lookFors) {
#	    if ($triple->getPredicate eq $self->{RDF_SCHEMA_URI}.$W3C::Rdf::RdfParser::PREDICATE_TYPE && $triple->getObject eq $lookFor){
	    if ($triple->getPredicate == $self->rdfToken($W3C::Rdf::RdfParser::PREDICATE_TYPE) && $triple->getObject == $lookFor){
		push (@ret, $triple->getSubject);
		next TRIPLE; # don't bother checking the remaining lookFors, we've already added the triple
	    }
	}
    }
    return @ret;
}

#####
# allPredicates (lookFors) - looks for all the nodes that have each
#                                     of the predicates listed in lookFors.
#         returns (an array of) two items:
#                 [0] - names of the nodes
#                 [1] - a "join" of all of the objects of the listed
#                       predicates.
#
# This function compensates for the lack of a syntax check on parameters passed
# to the RDF DB. A language devoted to a particular application would provide a
# scenario where a parser could say tha any node with proertyX is invalid wihtout proerties Y and Z.
#
# given DB triples:
#   triple('http://www.w3.org/TR/WD-rdf-syntax#type','B1','http://www.w3.org/TR/WD-rdf-syntax#Bag').
#   triple('http://www.w3.org/schema/certHTMLv1/access','tom','acc1').
#   triple('http://www.w3.org/TR/WD-rdf-syntax#predicate','genid1','http://www.w3.org/schema/certHTMLv1/access').
#   triple('http://www.w3.org/TR/WD-rdf-syntax#subject','genid1','tom').
#   triple('http://www.w3.org/TR/WD-rdf-syntax#object','genid1','acc1').
#   triple('http://www.w3.org/TR/WD-rdf-syntax#type','genid1','http://www.w3.org/TR/WD-rdf-syntax#Statement').
#   triple('http://www.w3.org/schema/certHTMLv1/accessor','tom','http://user1').
#   triple('http://www.w3.org/TR/WD-rdf-syntax#predicate','genid2','http://www.w3.org/schema/certHTMLv1/accessor').
#   triple('http://www.w3.org/TR/WD-rdf-syntax#subject','genid2','tom').
#   triple('http://www.w3.org/TR/WD-rdf-syntax#object','genid2','http://user1').
#   triple('http://www.w3.org/TR/WD-rdf-syntax#type','genid2','http://www.w3.org/TR/WD-rdf-syntax#Statement').
#   triple('http://www.w3.org/TR/WD-rdf-syntax#_1','B1','tom').
#   triple('http://www.w3.org/schema/certHTMLv1/access','joe','acc2').
#   triple('http://www.w3.org/TR/WD-rdf-syntax#predicate','genid3','http://www.w3.org/schema/certHTMLv1/access').
#   triple('http://www.w3.org/TR/WD-rdf-syntax#subject','genid3','joe').
#   triple('http://www.w3.org/TR/WD-rdf-syntax#object','genid3','acc2').
#   triple('http://www.w3.org/TR/WD-rdf-syntax#type','genid3','http://www.w3.org/TR/WD-rdf-syntax#Statement').
#   triple('http://www.w3.org/schema/certHTMLv1/accessor','joe','http://user2').
#   triple('http://www.w3.org/TR/WD-rdf-syntax#predicate','genid4','http://www.w3.org/schema/certHTMLv1/accessor').
#   triple('http://www.w3.org/TR/WD-rdf-syntax#subject','genid4','joe').
#   triple('http://www.w3.org/TR/WD-rdf-syntax#object','genid4','http://user2').
#   triple('http://www.w3.org/TR/WD-rdf-syntax#type','genid4','http://www.w3.org/TR/WD-rdf-syntax#Statement').
#   triple('http://www.w3.org/TR/WD-rdf-syntax#_2','B1','joe').
#   triple('http://www.w3.org/TR/WD-rdf-syntax#_3','B1','genid5').
#   triple('http://www.w3.org/schema/certHTMLv1/hasAccessTo','tom','http://resource/a').
#   triple('http://www.w3.org/schema/certHTMLv1/hasAccessTo','joe','http://resource/a').
#   triple('http://www.w3.org/schema/certHTMLv1/hasAccessTo','tom','http://resource/b').
#   triple('http://www.w3.org/schema/certHTMLv1/hasAccessTo','joe','http://resource/b').
# query for: 
#   ('http://www.w3.org/schema/certHTMLv1/access', 
#    'http://www.w3.org/schema/certHTMLv1/accessor', 
#    'http://www.w3.org/schema/certHTMLv1/hasAccessTo');
# we get:
#   ('tom', 'joe')

sub allPredicates {
    my ($self, $view, $lookFors) = @_;
    $view ||= $self->{TRIPLES};

    # store the name of each node (may be a genid) that has any of the appropriate predicates
    my %contenders;
    # foreach triple in the database
    foreach my $triple (@$view) {
	# store triples matching the lookFor predicats in arrays bin'd by the subject of those drawers
	map {push (@{$contenders{$triple->getSubject}[$_]}, $triple) if ($triple->getPredicate == $lookFors->[$_])} 0..$#$lookFors;
    }
    # you can review the resulting data structure with:
    # map {my $id = $_; print $id."\n";
    #      map {my $i = $_; map {print '  '.$i.' '.$_->show."\n"} @{$contenders{$id}[$i]};}
    #      0..$#{$contenders{$id}};} keys %contenders;
    # tom
    #   0 triple('http://www.w3.org/schema/certHTMLv1/access','tom','acc1').
    #   1 triple('http://www.w3.org/schema/certHTMLv1/accessor','tom','http://user1').
    #   2 triple('http://www.w3.org/schema/certHTMLv1/hasAccessTo','tom','http://resource/a').
    #   2 triple('http://www.w3.org/schema/certHTMLv1/hasAccessTo','tom','http://resource/b').
    # joe
    #   0 triple('http://www.w3.org/schema/certHTMLv1/access','joe','acc2').
    #   1 triple('http://www.w3.org/schema/certHTMLv1/accessor','joe','http://user2').
    #   2 triple('http://www.w3.org/schema/certHTMLv1/hasAccessTo','joe','http://resource/a').
    #   2 triple('http://www.w3.org/schema/certHTMLv1/hasAccessTo','joe','http://resource/b').

    # now extract the ones that have an entry for _each_ of the required predicates
    my (@nodes);
  NODE:
    foreach my $node (keys %contenders) {
	for (my $i = 0; $i <= $#$lookFors; $i++) {
	    my $triples = $contenders{$node}[$i];
	    next NODE if ($#$triples < $[);
	}
	push (@nodes, $node);
    }

    return @nodes; #, [$self->join(\@nodes, $lookFors)]);
}

#####
# join - build an array of all the cobinations of the objects for the sought predicates of the listed nodes 
#        I guess this is an n-dimensional matrix multiplication, but I haven't thougth about it much. I bet
#        my algorythms are pretty efficient so I won't bother chasing down a linear algebra cookbook.
#
# given the triple database and the predicate list shown above in the example for allPredicates and
# a list of nodes:
#   ('tom', 'joe')
# we get:
# .../certHTMLv1/access  .../certHTMLv1/accessor  .../certHTMLv1/hasAccessTo:
#	acc1			http://user1		http://resource/a
#	acc1			http://user1		http://resource/b
#	acc2			http://user2		http://resource/a
#	acc2			http://user2		http://resource/b

sub join {
    my ($self, $view, $nodes, $lookFors) = @_;
    $view ||= $self->{TRIPLES};

    # build a table of nodeValues containing each of the objects for the sought predicates of the listed nodes 
    my (%nodeValues); # nodeValues{node}{lookFor index} = [array of values]
    foreach my $node (@$nodes) { # walk nodes ...
	for (my $i = 0; $i <= $#$lookFors; $i++) { # and the lookFors ...
	    foreach my $triple ($self->triplesMatching($view, [[$lookFors->[$i], $node, undef]])) {
		# add each entry with this predicate for this node
		push (@{$nodeValues{$node}[$i]}, $triple->getObject);
		# print "$node $i ".$triple->getObject."\n";
	    }
	}
    }

    # build the array of all possible combinations of the values gathered above
    my @join;
    foreach my $node (@$nodes) {
	# The join is essentially a counting opperation where each digit is in a different base.
	# initialize the digit indexes to their first state
	my @indexes;
	map {push (@indexes, 0)} 0..$#$lookFors; # array of 0s as wide as @$lookFors
	for (my $done = 0; !$done; ) {
	    # build element array of the value for the curent index in each digit
	    my @element;
	    map {push (@element, $nodeValues{$node}[$_][$indexes[$_]])} 0..$#$lookFors;
	    # add with carry starting with the least significant digit
	    for (my $i = $0; $i <= $#$lookFors; $i++) {
		# if there's room left in the value array of the current digit
		if ($indexes[$i] < $#{$nodeValues{$node}[$i]}) {
		    # increment the current digit and push a new element
		    $indexes[$i]++;
		    last;
		} else {
		    # need to carry, so reset current current digit index to 0
		    if ($i < $#$lookFors) {
			$indexes[$i] = 0;
		    } else {
			# we're done if we try to carry past the number of digits we are counting
			$done = 1;
		    }
		}
	    }
	    # add the new element to the return array
	    push (@join, \@element);
	}
    }
    return @join;
}

sub members {
    my ($self, $view, $bagIds) = @_;
    $view ||= $self->{TRIPLES};
    my @ret;
    foreach my $bag ($self->anySubjects($view, $bagIds)) {
	foreach my $member ($self->anySubjects($view, [$bag->getObject])) {
	    push (@ret, $member);
	}
    }
    return @ret;
}

sub show {
    my ($self, $view, $prefix, $namespaceHandler) = @_;
    $prefix = '  ' if (!defined $prefix);

    my $ret = '';
    $ret .= $prefix.$self."\n";
    $ret .= $prefix.'RDF_PREFIX: '.$self->{RDF_PREFIX}."\n";
    $ret .= $prefix.'CURRENT_SOURCE_ATTRIBUTION: '.$self->{CURRENT_SOURCE_ATTRIBUTION}->dump."\n" if ($self->{CURRENT_SOURCE_ATTRIBUTION});
#    $ret .= $prefix.'CURRENT_PROPERTY_ATTRIBUTION: '.$self->{CURRENT_PROPERTY_ATTRIBUTION}->dump."\n";
    $ret .= $prefix."NAMESPACES:\n";map {$ret .= $prefix.'  '.$_.': '.$self->{NAMESPACES}{$_}."\n"} keys %{$self->{NAMESPACES}};
    $ret .= $prefix."CONTAINERS:\n";map {$ret .= $self->{CONTAINERS}{$_}->show($prefix.'  ')} keys %{$self->{CONTAINERS}};
    $ret .= $prefix."ABOUT_EACH:\n";map {$ret .= $prefix.'  '.$_.': ('.CORE::join(',', @{$self->{ABOUT_EACH}{$_}}).")\n"} keys %{$self->{ABOUT_EACH}};
    $ret .= $prefix."ABOUT_EACH_PREFIX:\n";map {$ret .= $prefix.'  '.$_.': ('.CORE::join(',', @{$self->{ABOUT_EACH_PREFIX}{$_}}).")\n"} keys %{$self->{ABOUT_EACH_PREFIX}};
    $ret .= $prefix."TRIPLES:\n";
    $ret .= CORE::join ("\n", $self->expandStatements($view, $prefix.'  ', $namespaceHandler))."\n";
    return $ret;
}

sub expandStatements {
    my ($self, $view, $prefix, $namespaceHandler, $flags) = @_;
    my @statements = $self->triplesMatching($view, [[undef, undef, undef]], $flags);
    return map {$_->show($prefix, $namespaceHandler, $flags)} @statements;
}

1;


__END__

=head1 NAME

W3C::Rdf::RdfDB - database to store RDF triples

=head1 SYNOPSIS

  use W3C::Rdf::RdfDB;
  my $rdfDB = new W3C::Rdf::RdfDB;
  my $desc = $rdfDB->addDescription(...);
  $desc->addPredicateObject($predicate, $object);
  $rdfDB->showTriples;

=head1 DESCRIPTION

Database to store RDF triples and a bunch of query opperations to retrieve them.

This module is part of the W3C::Rdf CPAN module.

=head1 AUTHOR

Eric Prud\'hommeaux <eric@w3.org>

=head1 SEE ALSO

W3C::Rdf::RdfParser(3) perl(1).

=cut
