#Copyright Massachusetts Institute of technology, 1998.
#Written by Eric Prud'hommeaux for the World Wide Web Consortium

use strict;
#use Exporter ();
require Exporter;
#require AutoLoader;

use W3C::SAX::HandlerBase;
use W3C::SAX::SAXParseException;
#use W3C::Rdf::MysqlDB qw($NodeIdType_ID $NodeIdType_Ref $NodeIdType_Gen $NodeIdType_Anon $NodeIdType_String);

$W3C::Rdf::RdfParser::REVISION = '$Id: RdfParser.pm,v 1.42 1999/12/14 13:14:40 eric Exp $ ';

package W3C::Rdf::RdfParser;
use vars qw($VERSION $DSLI @ISA @EXPORT @EXPORT_OK);
@ISA = qw(W3C::SAX::HandlerBase Exporter AutoLoader);
use W3C::Rdf::RdfDB; # qw(EATME);
# use W3C::Rdf::MysqlObjects::Containers qw(string_to_containerType);
# my %a = %W3C::Rdf::MysqlObjects::Containers::string_to_containerType;
@EXPORT = qw ($MODE_WAIT_BOGUS $MODE_WAIT_RDF $MODE_OBJECT $MODE_VALUE $MODE_PROPERTYELT $MODE_LITTERAL 
	      $MODE_ByPredicate $MODE_BySubjectType $MODE_BySubject $MODE_FAKEENTRY);
@EXPORT_OK = qw();
$VERSION = 0.94;
$DSLI = 'adpO';

use vars qw($MODE_WAIT_BOGUS $MODE_WAIT_RDF $MODE_OBJECT $MODE_VALUE $MODE_PROPERTYELT $MODE_LITTERAL 
	    $MODE_ByPredicate $MODE_BySubjectType $MODE_BySubject $MODE_FAKEENTRY);
($MODE_WAIT_BOGUS, $MODE_WAIT_RDF, $MODE_OBJECT, $MODE_VALUE, $MODE_PROPERTYELT, $MODE_LITTERAL, 
 $MODE_ByPredicate, $MODE_BySubjectType, $MODE_BySubject, $MODE_FAKEENTRY) = (0..10);
%W3C::Rdf::RdfParser::MODE_STRINGS = ($MODE_WAIT_BOGUS => 'BOGUS', 
		    $MODE_WAIT_RDF => 'WAIT_RDF', 
		    $MODE_OBJECT => 'OBJECT', 
		    $MODE_VALUE => 'VALUE', 
		    $MODE_PROPERTYELT => 'PROPERTYELT', 
		    $MODE_LITTERAL => 'LITTERAL', 
		    $MODE_ByPredicate => 'ByPredicate', 
		    $MODE_BySubjectType => 'BySubjectType', 
		    $MODE_BySubject => 'BySubject', 
		    $MODE_FAKEENTRY => 'FAKEENTRY');

#####
# Externally available constants

*W3C::Rdf::RdfParser::PREDICATE_TYPE = \ 'type';
*W3C::Rdf::RdfParser::PREDICATE_TYPE_LI = \ 'li';
*W3C::Rdf::RdfParser::PREDICATE_PREDICATE = \ 'predicate';
*W3C::Rdf::RdfParser::PREDICATE_SUBJECT = \ 'subject';
*W3C::Rdf::RdfParser::PREDICATE_OBJECT = \ 'object';

*W3C::Rdf::RdfParser::OBJECT_ALT = \ 'Alt';
*W3C::Rdf::RdfParser::OBJECT_BAG = \ 'Bag';
*W3C::Rdf::RdfParser::OBJECT_SEQ = \ 'Seq';
*W3C::Rdf::RdfParser::OBJECT_STATEMENT = \ 'Statement';

*W3C::Rdf::RdfParser::ATTRIBUTION_REIFICATION = \ '<RDF reification>';
*W3C::Rdf::RdfParser::ATTRIBUTION_PROPERTY = \ '<RDF property>';

# Tiny class to implement TripleObserver for W3C::Rdf::RdfParser's trust updates. This
# chains to W3C::Rdf::RdfParser's notifyTrustTriple. W3C::Rdf::RdfParser could be it's own
# TripleObserver, but it would be limited to a single notifyTriple functions
# in W3C::Rdf::RdfParser.

package W3C::Rdf::TrustObserver;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);
@ISA = qw(Exporter AutoLoader);

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $self = {};
    bless ($self, $class);
    ($self->{RDF_PARSER}) = @_;
    return $self;
}

sub notifyTriple {
    my ($self) = (shift);
    return $self->{RDF_PARSER}->notifyTrustTriple(@_);
}

package W3C::Rdf::ParseState;

sub new {
    my ($proto, $parser, $rdfDB, $tag) = @_;
    my $class = ref($proto) || $proto;
    my $self = {};
    bless ($self, $class);
    ($self->{RDF_PARSER}, $self->{-rdfDB}, $self->{TAG}) = ($parser, $rdfDB, $tag);
    return $self;
}

sub getMode ($) {my ($self) = @_; return $self->{MODE};}
sub setMode ($$) {my ($self, $mode) = @_; $self->{MODE} = $mode;}
sub getInvalid ($) {my ($self) = @_; return $self->{INVALID};}
sub setInvalid ($$) {my ($self, $invalid) = @_; $self->{INVALID} = $invalid;}
sub getTag ($) {my ($self) = @_; return $self->{TAG};}
sub getNode ($) {my ($self) = @_; return $self->{NODE};}
sub setNode ($$) {my ($self, $node) = @_; $self->{NODE} = $node;}
sub getContainer ($) {my ($self) = @_; return $self->{CONTAINER};}
sub setContainer ($$) {my ($self, $container) = @_; $self->{CONTAINER} = $container;}
sub getPredicate ($) {my ($self) = @_; return $self->{PREDICATE};}
sub setPredicate ($$) {my ($self, $predicate) = @_; $self->{PREDICATE} = $predicate;}
sub getReification ($) {my ($self) = @_; return $self->{REIFICATION};}
sub setReification ($$) {my ($self, $reification) = @_; $self->{REIFICATION} = $reification;}
sub getReificationBag ($) {my ($self) = @_; return $self->{REIFICATION_BAG};}
sub setReificationBag ($$) {my ($self, $reificationBag) = @_; $self->{REIFICATION_BAG} = $reificationBag;}
sub getAttribution ($) {my ($self) = @_; return $self->{ATTRIBUTION};}
sub setAttribution ($$) {my ($self, $attribution) = @_; $self->{ATTRIBUTION} = $attribution;}
sub getSum ($) {my ($self) = @_; return $self->{SUM};}
sub nextIsSum () {my ($self) = @_; $self->{NEXT_IS_SUM} = 1;}
sub addSumString ($) {my ($self, $addendum) = @_; $self->{SUM} ^= unpack ('%16C*', $addendum) if (!$self->{NEXT_IS_SUM});}
sub addPreSum ($) {my ($self, $addendum) = @_; $self->{SUM} ^= $addendum if (!$self->{NEXT_IS_SUM});}

sub showHeader (;$$) {
    my ($prefix, $wd) = @_;
    my $ret;
    $ret = sprintf($prefix."%10s %10s %12s %${wd}s %${wd}s %${wd}s %${wd}s %${wd}s %30s",
		   'self', 'tag', 'mode', 'parser', 'db', 'attribution', 'node', 
		   'predicate', 'container');
    return $ret;
}

sub show ($;$$) {
    my ($self, $prefix, $wd) = @_;
    my $ret;
    my $node = (defined $self->{NODE}) ? 
	'('.substr($self->{NODE}->show($self->{RDF_PARSER}), -$wd).')' : 
	    '';
    my $predicate = ($self->{RDF_PARSER} && $self->{RDF_PARSER}->{-namespaceHandler}) ? 
	substr($self->{RDF_PARSER}->{-namespaceHandler}->unmapNamespace($self->{PREDICATE}), -$wd) :
	    '';
    $ret = sprintf($prefix."%10s %10s %12s %-${wd}s %-${wd}s %-${wd}s %-${wd}s %-${wd}s %-30s",
		   $self=~m/\(([a-f0-9x]+)\)/, 
		   substr($self->{TAG}, -10), $W3C::Rdf::RdfParser::MODE_STRINGS{$self->{MODE}}, 
		   substr($self->{RDF_PARSER}, -$wd), 
		   substr($self->{-rdfDB}, -$wd), 
		   substr($self->{ATTRIBUTION}, -$wd), 
		   substr($node, -$wd), 
		   substr($predicate, -$wd),
		   substr($self->{CONTAINER}, -30));
    return $ret;
}

package W3C::Rdf::RdfParser;

#####
# new - prepare a W3C::Rdf::RdfParser with a new connection

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;

    # default to a trusting state with no trusted or untrusted ids
    my ($parms) = @_;

    my $self = $class->SUPER::new();
    bless ($self, $class);
    $self->{NEXT_ASSIGNED_ID} = 1;
    $self->{-nextTrustControlBlock} = defined $parms->{-nextTrustControlBlock} ? $parms->{-nextTrustControlBlock} : 1;
    $self->{-trustedIds} = defined $parms->{-trustedIds} ? $parms->{-trustedIds} : [];
    $self->{-untrustedIds} = defined $parms->{-untrustedIds} ? $parms->{-untrustedIds} : [];
    $self->{-rdfDB} = $parms->{-rdfDB};
    $self->{-namespaceHandler} = $parms->{-namespaceHandler};
    $self->{-errorHandler} = $parms->{-errorHandler};
    $self->{-appendAttribution} = $parms->{-appendAttribution};

    # these should probably be passed or set parameters
    $self->{DEFAULT_DOC_TRUST} = undef;
    $self->{DEFAULT_LINK_TRUST} = undef;

    $self->{TRUST_SCHEMA_URI} = 'http://www.w3.org/RDF/schema/Trust/';
    $self->{TRUST_OBSERVER} = new W3C::Rdf::TrustObserver($self); # friend class
    $self->{STATE_STACK} = [];
    $self->{RDF_DECL} = 0;
    return $self;
}

sub append ($) { # prepare for subsequent XML documents
    $_[0]->{RDF_DECL} = -1;
}

sub setState ($) {
    my ($self, $mode) = @_;
    $self->{MODE} = $mode;
}

sub setNamespaceHandler {
    my ($self, $namespaceHandler) = @_;
    $self->{-namespaceHandler} = $namespaceHandler;
}

sub addNamespace {
    my ($self, $sAs, $sHref) = @_;
    return $self->{-rdfDB}->addNamespace($sAs, $sHref);
}

# support the W3C::SAX::NamespaceHandler::unmapNamespace function so we can _be_ a W3C::SAX::NamespaceHandler

sub unmapNamespace {
    my ($self, $name) = @_;
    warn 'W3C::Rdf::RdfParser::unmapNamespace has been deprecated';
    require ('perl5db.pl');
    &DB::dump_trace;
    die;
    $name = $self->{-namespaceHandler}->unmapNamespace($name, $self->{URI_STRING}) if ($self->{-namespaceHandler});
    $name =~ s/\A\Q$self->{URI_STRING}\E/<source>/;
    return $name;
}

sub newRdfParserContext {
    my ($self) = @_;
    for (my $i = 0; $i < @{$self->{STATE_STACK}} -1; $i++) {
	$self->{STATE_STACK}[$i]->{MODE} = $MODE_WAIT_RDF;
    }
}

sub setRdfSchemaURI {
    my ($self, $schemaURI) = @_;
    $self->{PARSER}->{RDF_SCHEMA_URI} = $schemaURI;
}

sub getRdfPrefix {
    my ($self) = @_;
    return $self->{-rdfDB}->getRdfPrefix;
}

sub getSystemId {
    my ($self) = @_;
    return $self->{SYSTEM_ID};
}

sub getRootAttribution {
    my ($self) = @_;
    return $self->{ROOT_ATTRIBUTION};
}

sub getRdfDB {
    my ($self) = @_;
    return $self->{-rdfDB};
}

# Parsing support functions dealing with the TAG_STATE

# maintain state in a stack - could be a first class object, but I don't see much point

sub notifyTrustTriple {
    my ($self, $triple) = @_;
#    return if (!$triple->trusted); # @@@ maybe that should go into an W3C::Rdf::RdfDB::triplesMatching parameter
    my $object = $triple->object;
    if ($triple->predicate eq $self->{TRUST_SCHEMA_URI}.'trust') {
	# figure out what to do if it's a bag...
	$self->{-trustedIds}{$object} = $self->{-rdfDB}->getAttribution($AttributionType_SOURCE, $object, $self->{ATTRIBUTION}, undef, 1) 
	    if (!exists $self->{-trustedIds}{$object});
    } elsif ($triple->predicate eq $self->{TRUST_SCHEMA_URI}.'disTrust') {
	$self->{-trustedIds}{$object} = $self->{-rdfDB}->getAttribution($AttributionType_SOURCE, $object, $self->{ATTRIBUTION}, undef, 0) 
    } elsif ($triple->predicate eq $self->{TRUST_SCHEMA_URI}.'defaultDocTrust') {
	$self->{DEFAULT_DOC_TRUST} = $object eq 'trust';
    } elsif ($triple->predicate eq $self->{TRUST_SCHEMA_URI}.'defaultLinkTrust') {
	$self->{DEFAULT_LINK_TRUST} = $object eq 'trust';
    } elsif ($triple->predicate eq $self->{TRUST_SCHEMA_URI}.'sum') {
	$self->{STATE_STACK}[-1]->getAttribution->setStatedSum($object);
    }
#    print $triple->show('', $self)."\n";
}

sub checkTrust {
    my ($self, $id) = @_;
    if (my $nextAttribution = $self->{-trustedIds}{$id}) {
	$self->{ATTRIBUTION} = $nextAttribution;
    }
    return $self->{ATTRIBUTION};
}

# add ordinals to Container if we are in an <li> element

#####
# documentHandler routines

sub setDocumentLocator {
    my ($self, $locator) = @_;
    $self->{DOCUMENT_LOCATOR} = $locator;
    $self->{URI_STRING} = $locator->getPublicId;
    $self->{SYSTEM_ID} = $self->{-rdfDB}->getUri($self->{URI_STRING});
    $self->{ROOT_ATTRIBUTION} = $self->{-rdfDB}->getAttribution($AttributionType_SOURCE, $self->{SYSTEM_ID}, undef, undef, $self->{-nextTrustControlBlock}, !$self->{-appendAttribution});
    $self->{ATTRIBUTION} = $self->{ROOT_ATTRIBUTION};
    $self->{-rdfDB}->setSource($self->{URI_STRING});
    # $$$ assume that setDocumentLocator means we have started a new document? What about recursing back up?
    $self->{-rdfDB}->setSourceAttribution($self->{ATTRIBUTION});
}

sub startDocument {
    my ($self) = @_;
    $self->SUPER::startDocument();
    $self->newRdfParserContext;
    my $nextAttribution = $self->{-rdfDB}->getSourceAttribution();

    # set up a tag state that defaults to parsing RDF in case it is not in an :RDF element
    $self->{MODE} = $MODE_OBJECT if (!defined $self->{MODE}); # ref M&S#@@@

    $self->{-rdfDB}->registerTripleObserver([[$self->{TRUST_SCHEMA_URI}.'trust', undef, undef], 
					     [$self->{TRUST_SCHEMA_URI}.'disTrust', undef, undef], 
					     [$self->{TRUST_SCHEMA_URI}.'defaultDocTrust', undef, undef], 
					     [$self->{TRUST_SCHEMA_URI}.'defaultLinkTrust',undef, undef], 
					     [$self->{TRUST_SCHEMA_URI}.'sum',undef, undef]],$self->{TRUST_OBSERVER});
}

sub endDocument {
    my ($self) = @_;
    # flush top attribution 'cause checksum has changed
    $self->{ROOT_ATTRIBUTION}->flush({-discretionaryUpdate => 1, -default => {'parent' => 1}});
    $self->SUPER::endDocument();
}

sub pushParseState () {
    my ($self, $tag) = @_;
    my $lastParseState = $self->{STATE_STACK}[-1];
    my $nextParseState = new W3C::Rdf::ParseState($self, $self->{-rdfDB}, $tag);
    $nextParseState->setAttribution($self->{ATTRIBUTION});
    push (@{$self->{STATE_STACK}}, $nextParseState);
    return ($lastParseState, $nextParseState);
}

sub popParseState () {
    my ($self) = @_;
    my $lastParseState = pop (@{$self->{STATE_STACK}});
    if (@{$self->{STATE_STACK}} > 0) {
	my $nextParseState = $self->{STATE_STACK}[-1];
	my $nextAttribution = $nextParseState->getAttribution;
	if ($self->{ATTRIBUTION} != $nextAttribution) {
	    $self->{ATTRIBUTION}->flush;
	    $self->{ATTRIBUTION} = $nextParseState->getAttribution;
	}
	$self->{MODE} = $nextParseState->getMode;
    } else {
	$self->{ATTRIBUTION} = $self->{ROOT_ATTRIBUTION};
	$self->{MODE} = $MODE_WAIT_RDF;
    }
}

sub showParseState ($;$$) {
    my ($self, $prefix, $wd) = @_; $prefix = '' if (!defined $prefix); $wd ||= 20;
    my $ret = &W3C::Rdf::ParseState::showHeader($prefix, $wd)."\n";
    map {$ret .= $_->show($prefix, $wd)."\n"} @{$self->{STATE_STACK}};
    return $ret;
}

# some imple notes that need to be gathered and well-presented
#			reification	reificationBag
# 0	D1		I-1		B-2
# 1	D0.attr-value	gen		B0
# 2	P0.resource	I0		B-1
# 3	P0.attr-value	gen+/I0*	B0+/B1*
#			+ - nested description node
#			* - attribute-value triples

sub ensureReification ($) {
    my ($self, $depth) = @_;
    my $ret = undef;
    $ret = $self->{STATE_STACK}[$depth-1]->getReification if (defined $depth);
    $ret = $self->{-rdfDB}->createGenId($self->{ATTRIBUTION}) if (!defined $ret);
    $self->{STATE_STACK}[$depth-1]->setReification($ret) if (defined $depth);
    return $ret;
}

sub ensureReificationBag ($) {
    my ($self, $depth) = @_;
    my $ret = undef;
    $ret = $self->{STATE_STACK}[$depth-1]->getReificationBag if (defined $depth);
    $ret = $self->{-rdfDB}->createGenId($self->{ATTRIBUTION}) if (!defined $ret);
    $self->{STATE_STACK}[$depth-1]->setReificationBag($ret) if (defined $depth);
    return $ret;
}

sub checkStackState ($) {
    my ($self, $depth) = @_;
    return (defined $self->{STATE_STACK}[$depth] && 
	    $self->{STATE_STACK}[$depth]->getMode != $MODE_WAIT_RDF && 
	    !$self->{STATE_STACK}[$depth]->getInvalid);
}

# general use productions not worth listing each place they come up:
# [6.19] Qname          ::= [ NSprefix ':' ] name
# [6.22] name           ::= (any legal XML name symbol)
# [6.20] URI-reference  ::= string, interpreted per [URI]
# [6.21] IDsymbol       ::= (any legal XML name symbol)
# [6.23] NSprefix       ::= (any legal XML namespace prefix)
# [6.24] string         ::= (any XML text, with "<", ">", and "&" escaped)

sub startElement {
    my ($self, $name, $attributeList) = @_;
    my ($lastParseState, $nextParseState) = $self->pushParseState($name); # print $self->showParseState(':', 25)."\n";
    my $rdfPrefix = $self->getRdfPrefix; die if (!defined $self->{-namespaceHandler});
    if ($self->{-namespaceHandler} && $self->{-namespaceHandler}->can('getUnknownNamespaces') && 
	grep {/^rdf$/} $self->{-namespaceHandler}->getUnknownNamespaces) {
	$rdfPrefix = 'rdf:';
    }
    if ($name eq $self->{TRUST_SCHEMA_URI}.'sum') {
	$nextParseState->nextIsSum;
    } else {
	$nextParseState->addSumString($name);
	$nextParseState->addPreSum($attributeList->sum([]));
    }
    if ($self->{MODE} == $MODE_ByPredicate || 
	$self->{MODE} == $MODE_BySubjectType || 
	$self->{MODE} == $MODE_BySubject) {
	# ($self->{MODE} == $MODE_WAIT_RDF) { # shortcut out when we're not parsing
	my ($id, $about, $aboutEach, $aboutEachPrefix, $reification, $bagId, $type, $parseType) = 
	    ($attributeList->getOne($rdfPrefix.'ID'), 
	     $attributeList->getOne($rdfPrefix.'about'), 
	     $attributeList->getOne($rdfPrefix.'aboutEach'), 
	     $attributeList->getOne($rdfPrefix.'aboutEachPrefix'), 
	     $attributeList->getOne($rdfPrefix.'reification'), 
	     $attributeList->getOne($rdfPrefix.'bagID'), 
	     $attributeList->getOne($rdfPrefix.'type'), 
	     $attributeList->getOne($rdfPrefix.'parseType'));
	$self->{-rdfDB}->absoluteLink($self->{SYSTEM_ID}, \$about, \$aboutEach, \$aboutEachPrefix, \$reification, \$bagId, \$type);
	$self->{-rdfDB}->absoluteName($self->{SYSTEM_ID}, \$bagId);
	$bagId = $self->{-rdfDB}->createGenId($self->{ATTRIBUTION}) if (!$bagId);
	if ($about) {
	    $id = $about;
	} elsif (!defined $aboutEach && !defined $aboutEachPrefix) {
	    $self->{-rdfDB}->absoluteName($self->{SYSTEM_ID}, \$id);
#	    if ($id) {   # @@@ is this handled by createDescription?
#		$self->{-rdfDB}->absoluteName($self->{SYSTEM_ID}, \$id);
#	    } else {
#		$id = $self->{-rdfDB}->createGenId($self->{ATTRIBUTION});
#	    }
	}
	$self->checkTrust($id);

	my ($description, $node) = 
	    $self->{-rdfDB}->createDescription($type, $id, $aboutEach, 
							   $aboutEachPrefix, $bagId, $self->{ATTRIBUTION});
	$nextParseState->setNode($node);
	$nextParseState->setReification($reification);
	$nextParseState->setReificationBag($bagId);

	# triple asserted by the existence of this node within a propertyElt
	if ($self->checkStackState(-2)) {
#	    my $predicate = $self->{STATE_STACK}[-1]->getPredicate;
	    my $predicate = $name;
	    $self->{-rdfDB}->absoluteLink($self->{SYSTEM_ID}, \$predicate);
	    my $subject = $self->{STATE_STACK}[-2]->getNode;
	    my $object = $node;

	    if (ref $predicate && ref $subject && ref $object) {
	    #			reification	reificationBag
	    # 0	D1		I-1		B-2
		my ($reification, $container) = ($self->ensureReification(0), $self->ensureReificationBag(0)); # @@@ can't decide
		my $statement = $self->{-rdfDB}->addStatement ($predicate, $subject, $object, 
								       $container, $reification, $self->{ATTRIBUTION});
	    }
#	    print $statement->show(':', $self, {-showIds => 0, -showTypes => 0})."\n";
	}

	# triples asserted by attribute="value"
	for (my $i = 0; $i < $attributeList->getLength; $i++) {
	    my ($attributeName, $attributeValue) = ($attributeList->getName($i), $attributeList->getValue($i));
	    if ($attributeName !~ m/\A \Q$rdfPrefix\E (ID|about|aboutEach|aboutEachPrefix|reification|agID|type|parseType) \Z/x) {
		# [6.10] propAttr       ::= typeAttr
		#                         | propName '="' string '"' (with embedded quotes escaped)
		# [6.11] typeAttr       ::= 'type="' URI-reference '"' @@@
		#			reification	reificationBag
		# 1	D0.attr-value	gen		B0
		my $predicate = $attributeName;
		$self->{-rdfDB}->absoluteLink($self->{SYSTEM_ID}, \$predicate);
		my $subject = $node;
		my $object = $self->{-rdfDB}->getString($attributeValue);
		my ($reification, $container) = ($self->ensureReification(undef), $self->ensureReificationBag(0));
		my $statement = $self->{-rdfDB}->addStatement($predicate, $subject, $object, 
									     $container, $reification, $self->{ATTRIBUTION});
	    }
	}

	if ($name eq $rdfPrefix.'Statement') { # quoting primitave
	    $nextParseState->setPendingStatement();
	} else {
	}

	$nextParseState->setMode($self->{MODE} = ($parseType eq 'RDF') ? $MODE_OBJECT : $MODE_ByPredicate);
    } else { # shortcut out when we're not parsing
	if ($name eq $rdfPrefix.'RDF') {
	    #  [6.1] RDF            ::= '<rdf:RDF>' obj* '</rdf:RDF>'
	    if ($self->{RDF_DECL} == 1) { # 2nd or later RDF block
		if ($self->{MODE} != $MODE_WAIT_RDF) {
		    $self->{-errorHandler}->error(new W3C::SAX::SAXParseException ('Not sure how to deal with nested <'.
										   $rdfPrefix.'RDF>'.
									    ' so I thought I\'d make a pain in the ass of myself'. 
										   ':'.$self->{DOCUMENT_LOCATOR}->getPublicId));
		}
	    } else { # first RDF block found
		$self->{-nextTrustControlBlock} = 0;
		# @@@ Clear W3C::Rdf::RdfDB additions since the documentStart: triples, containers, aboutEach[Prefix].
		#     Haven't worked out how to do that (without complicating the interface), so I'll just clear the whole thing.
		$self->{-rdfDB}->clear if ($self->{RDF_DECL} != -1);
		$self->{RDF_DECL} = 1;
	    } # should new W3C::Rdf::SourceAttribution just happen on the first pass (immediately above here)?
	    $self->newRdfParserContext;
	    $nextParseState->setInvalid(1);
	    $nextParseState->setMode($self->{MODE} = $MODE_OBJECT);
	} elsif ($name eq $rdfPrefix.'ByPredicate') {
	    $nextParseState->setMode($self->{MODE} = $MODE_ByPredicate);
	} elsif ($name eq $rdfPrefix.'BySubjectType') {
	    $nextParseState->setMode($self->{MODE} = $MODE_BySubjectType);
	} elsif ($name eq $rdfPrefix.'BySubject') {
	    $nextParseState->setMode($self->{MODE} = $MODE_BySubject);
	} elsif ($self->{MODE} == $MODE_WAIT_BOGUS) { # make sure we never end up with a 0 for mode
	    die;
	} elsif ($self->{MODE} == $MODE_WAIT_RDF) { # ignore tags outside an RDF block
	    return;
	} elsif ($self->{MODE} == $MODE_OBJECT || $self->{MODE} == $MODE_VALUE) {
	    my ($type, $id, $bagId, $aboutEach, $aboutEachPrefix, $isContainer, $isDescription);
	    #  [6.2] obj            ::= description | container
	    if ($name =~ m/\A \Q$rdfPrefix\E (Seq|Bag|Alt) \Z/x) {
		#  [6.4] container      ::= sequence | bag | alternative
		# [6.25] sequence       ::= '<rdf:Seq' idAttr? '>' member* '</rdf:Seq>'
		#                         | '<rdf:Seq' idAttr? memberAttr* '/>'
		# [6.26] bag            ::= '<rdf:Bag' idAttr? '>' member* '</rdf:Bag>'
		#                         | '<rdf:Bag' idAttr? memberAttr* '/>'
		# [6.27] alternative    ::= '<rdf:Alt' idAttr? '>' member+ '</rdf:Alt>'
		#                         | '<rdf:Alt' idAttr? memberAttr* '/>'
		$id = $attributeList->getOne($rdfPrefix.'ID');
		$self->{-rdfDB}->absoluteName($self->{SYSTEM_ID}, \$id, \$bagId);
		$bagId = $self->{-rdfDB}->createGenId($self->{ATTRIBUTION}) if (!$bagId);
	    } else {
		#  [6.3] description    ::= '<rdf:Description' idAboutAttr? bagIdAttr? propAttr* '/>'
		#                         | '<rdf:Description' idAboutAttr? bagIdAttr? propAttr* '>'
		#                                propertyElt* '</rdf:Description>'
		#                         | typedNode
		if ($name eq $rdfPrefix.'Description') {
		    # new  object who's type not yet known
		    $isDescription = 1;
		} else {
		    # [6.13] typedNode      ::= '<' typeName idAboutAttr? bagIdAttr? propAttr* '/>'
		    #                         | '<' typeName idAboutAttr? bagIdAttr? propAttr* '>'
		    #                               propertyElt* '</' typeName '>'
		    # [6.15] typeName       ::= Qname
		    $type = $name;
		    $self->{-rdfDB}->absoluteLink($self->{SYSTEM_ID}, \$type);
		}
		#  [6.5] idAboutAttr    ::= idAttr | aboutAttr | aboutEachAttr
		#  [6.6] idAttr         ::= 'ID="' IDsymbol '"'
		#  [6.7] aboutAttr      ::= 'about="' URI-reference '"'
		#  [6.8] aboutEachAttr  ::= 'aboutEach="' URI-reference '"'
		#                         | 'aboutEachPrefix="' string '"'
		#  [6.9] bagIdAttr      ::= 'bagID="' IDsymbol '"'
		my $about;
		($id, $about, $aboutEach, $aboutEachPrefix, $bagId) = 
		    ($attributeList->getOne($rdfPrefix.'ID'), 
		     $attributeList->getOne($rdfPrefix.'about'), 
		     $attributeList->getOne($rdfPrefix.'aboutEach'), 
		     $attributeList->getOne($rdfPrefix.'aboutEachPrefix'), 
		     $attributeList->getOne($rdfPrefix.'bagID'));
		$self->{-rdfDB}->absoluteLink($self->{SYSTEM_ID}, \$about, \$aboutEach, \$aboutEachPrefix);
		$self->{-rdfDB}->absoluteName($self->{SYSTEM_ID}, \$bagId);
		$bagId = $self->{-rdfDB}->createGenId($self->{ATTRIBUTION}) if (!$bagId);
		if ($about) {
		    $id = $about;
		} elsif (!defined $aboutEach && !defined $aboutEachPrefix) {
		    if (!$id) {
			$id = $self->{-rdfDB}->createGenId($self->{ATTRIBUTION});
		    } else {
			$self->{-rdfDB}->absoluteName($self->{SYSTEM_ID}, \$id);
		    }
		}
		$self->checkTrust($id);
	    }

	    my ($description, $node) = 
		$self->{-rdfDB}->createDescription($type, $id, $aboutEach, 
						   $aboutEachPrefix, $bagId, $self->{ATTRIBUTION});
	    $nextParseState->setNode($node);
	    $nextParseState->setReificationBag($bagId) if (!$isContainer);

	    # triple asserted by the existence of this node within a propertyElt
	    if ($self->checkStackState(-2)) {
		my $predicate = $lastParseState->getPredicate;
		my $subject = $self->{STATE_STACK}[-2]->getNode;
		my $object = $node;

		#			reification	reificationBag
		# 0	D1		I-1		B-2
		my ($reification, $container) = ($self->ensureReification(-1), $self->ensureReificationBag(-2));
		my $statement = $self->{-rdfDB}->addStatement ($predicate, $subject, $object, 
									      $container, $reification, $self->{ATTRIBUTION});
#		print $statement->show(':', $self, {-showIds => 0, -showTypes => 0})."\n";
	    }

	    # triples asserted by attribute="value"
	    for (my $i = 0; $i < $attributeList->getLength; $i++) {
		my ($attributeName, $attributeValue) = ($attributeList->getName($i), $attributeList->getValue($i));
		if ($attributeName !~ m/\A \Q$rdfPrefix\E (ID|about|aboutEach|aboutEachPrefix|bagID) \Z/x) {
		    # [6.10] propAttr       ::= typeAttr
		    #                         | propName '="' string '"' (with embedded quotes escaped)
		    # [6.11] typeAttr       ::= 'type="' URI-reference '"' @@@
		    #			reification	reificationBag
		    # 1	D0.attr-value	gen		B0
		    my $predicate = $attributeName;
		    $self->{-rdfDB}->absoluteLink($self->{SYSTEM_ID}, \$predicate);
		    my $subject = $node;
		    my $object = $self->{-rdfDB}->getString($attributeValue);
		    my ($reification, $container) = ($self->ensureReification(undef), $self->ensureReificationBag(0));
		    my $statement = $self->{-rdfDB}->addStatement($predicate, $subject, $object, 
										 $container, $reification, $self->{ATTRIBUTION});
		}
	    }
	    $nextParseState->setMode($self->{MODE} = $MODE_PROPERTYELT);
	} elsif ($self->{MODE} == $MODE_PROPERTYELT) {
	    if ($lastParseState->getTag =~ m/\A \Q$rdfPrefix\E (Seq|Bag|Alt) \Z/x && $name !~ m/\A \Q$rdfPrefix\E (li) \Z/x) {
		# [6.28] member         ::= referencedItem | inlineItem
		# [6.29] referencedItem ::= '<rdf:li' resourceAttr '/>'
		# [6.30] inlineItem     ::= '<rdf:li>' value </rdf:li>'
		# [6.18] resourceAttr   ::= 'resource="' URI-reference '"'
		$self->{-errorHandler}->error(new W3C::SAX::SAXParseException ('expected <'.$rdfPrefix.
									       ':li> element in an RDF Container', 
									       $self->{DOCUMENT_LOCATOR}));
	    }
	    # [6.12] propertyElt    ::= '<' propName idAttr? '>' value '</' propName '>'
	    #                         | '<' propName idAttr? parseLiteral '>' @@@
	    #                               literal '</' propName '>'
	    #                         | '<' propName idAttr? parseResource '>' @@@
	    #                               propertyElt* '</' propName '>'
	    #                         | '<' propName idRefAttr? bagIdAttr? propAttr* '/>'
	    #x[6.12] propertyElt    ::= '<' propName idAttr? '>' value '</' propName '>'
	    #x                        | '<' propName idRefAttr? bagIdAttr? propAttr* '/>'
	    # [6.14] propName       ::= Qname
	    #  [6.6] idAttr         ::= 'ID="' IDsymbol '"'
	    # [6.17] value          ::= obj | string
	    # [6.16] idRefAttr      ::= idAttr | resourceAttr
	    # [6.18] resourceAttr   ::= 'resource="' URI-reference '"'
	    #  [6.9] bagIdAttr      ::= 'bagID="' IDsymbol '"'
	    my ($id, $resource, $bagId, $parseType, $statedSum) = ($attributeList->getOne($rdfPrefix.'ID'), 
								   $attributeList->getOne($rdfPrefix.'resource'), 
								   $attributeList->getOne($rdfPrefix.'bagID'), 
								   $attributeList->getOne($rdfPrefix.'parseType'), 
								   $attributeList->getOne($self->{TRUST_SCHEMA_URI}.'$sum'));
	    $self->{-rdfDB}->absoluteName($self->{SYSTEM_ID}, \$id, \$bagId); # , \$resource
	    my $predicate = $name;
	    $self->{-rdfDB}->absoluteLink($self->{SYSTEM_ID}, \$predicate, \$resource);
	    $nextParseState->setNode($lastParseState->getNode());
	    $nextParseState->setPredicate($predicate);
	    $nextParseState->setReification($id);
	    my $object = undef; # may be used if there are attribute-value pairs to triplize
	    if (defined $resource || (defined $parseType && ($parseType eq 'Resource' || $parseType eq 'ByPredicate'))) {
		#			reification	reificationBag
		# 2	P0.resource	I0		B-1
		my ($reification, $container) = ($self->ensureReification(0), $self->ensureReificationBag(-1));
		my $subject = $nextParseState->getNode;
		$object = $resource ? $resource : $self->{-rdfDB}->createGenId($self->{ATTRIBUTION});
		my $statement = $self->{-rdfDB}->addStatement($predicate, $subject, $object, 
									     $container, $reification, $self->{ATTRIBUTION});
#		print $statement->show(':', $self, {-showIds => 0, -showTypes => 0})."\n";
	    }

	    for (my $i = 0; $i < $attributeList->getLength; $i++) {
		my ($attributeName, $attributeValue) = ($attributeList->getName($i), $attributeList->getValue($i));
		if ($attributeName !~ m/\A \Q$rdfPrefix\E (ID|resource|bagID|parseType) \Z/x) {
		    # [6.10] propAttr       ::= typeAttr
		    #                         | propName '="' string '"' (with embedded quotes escaped)
		    # [6.11] typeAttr       ::= 'type="' URI-reference '"' @@@
		    #			reification	reificationBag
		    my $predicate = $attributeName;
		    $self->{-rdfDB}->absoluteLink($self->{SYSTEM_ID}, \$predicate);

		    # 3	P0.attr-value	gen+/I0*	B0+/B1*
		    #			+ - nested description node
		    #			* - attribute-value triples
		    if (!defined $object) {
			# prepare for this propAttr and future propAttrs
			$object = $self->{-rdfDB}->createGenId($self->{ATTRIBUTION});
		    }
		    my $innerObject = $self->{-rdfDB}->getString($attributeValue);
		    my ($reification, $container) = ($self->ensureReification(-1), $self->ensureReificationBag(-2));
		    my $statement = $self->{-rdfDB}->addStatement($predicate, $object, $innerObject, 
								  $container, $reification, $self->{ATTRIBUTION});
		}
	    }

	    if (defined $parseType && ($parseType eq 'Resource' || $parseType eq 'ByPredicate')) {
		# create fake stack entry with node $object
		$nextParseState->setMode($MODE_FAKEENTRY);
		($lastParseState, $nextParseState) = $self->pushParseState('fake entry');
		$nextParseState->setNode($object);
		$self->{MODE} = $parseType eq 'ByPredicate' ? $MODE_ByPredicate : $MODE_PROPERTYELT;
	    } else {
		$self->{MODE} = ($parseType eq 'Literal') ? $MODE_LITTERAL : $MODE_VALUE;
	    }
	    $nextParseState->setMode($self->{MODE});
	} elsif ($self->{MODE} == $MODE_LITTERAL) {
	    my $elementStr = $name;
	    for (my $i = 0; $i < $attributeList->getLength; $i++) {
		my ($attributeName, $attributeValue) = ($attributeList->getName($i), $attributeList->getValue($i));
		$elementStr .= ' '.$attributeName.'="'.$attributeValue.'"';
	    }
	    $self->{CURRENT_LITTERAL} .= '<'.$elementStr.'>';
	    $nextParseState->setMode($self->{MODE} = $MODE_LITTERAL);
	}
    }
    # print $self->showParseState(':', 25)."\n";
}
	    
sub endElement {
    my ($self, $name) = @_;
    if ($self->{STATE_STACK}[-1]->getMode == $MODE_FAKEENTRY) { # ouch!
	$self->{STATE_STACK}[-2]->addPreSum($self->{STATE_STACK}[-1]->getSum) if ($self->checkStackState(-2));
	$self->{ATTRIBUTION}->_setSum($self->{STATE_STACK}[-1]->getSum);
	$self->popParseState;
    }
    $self->{STATE_STACK}[-2]->addPreSum($self->{STATE_STACK}[-1]->getSum) if ($self->checkStackState(-2));
    $self->{ATTRIBUTION}->_setSum($self->{STATE_STACK}[-1]->getSum);
    if ($self->{MODE} != $MODE_WAIT_RDF) { # shortcut out when we're not parsing
	if ($self->{MODE} == $MODE_LITTERAL) {
	    if ($self->{STATE_STACK}[-2]->getMode == $MODE_LITTERAL) {
		$self->{CURRENT_LITTERAL} .= '</'.$name.'>';
	    } else {
		my $subject = $self->{STATE_STACK}[-2]->getNode;
		my $predicate = $self->{STATE_STACK}[-1]->getPredicate;
		my $object = $self->{-rdfDB}->getString($self->{CURRENT_LITTERAL});
		my ($reification, $container) = ($self->ensureReification(-1), $self->ensureReificationBag(-2));
		my $statement = $self->{-rdfDB}->addStatement($predicate, $subject, $object, 
									     $container, $reification, $self->{ATTRIBUTION});
	    }
	}
	$self->popParseState;
    }
}

sub characters {
    my ($self, $ch, $start, $length) = @_;
    $self->{STATE_STACK}[-1]->addSumString($ch) if ($self->checkStackState(-1));
    my $shrunk = $ch;
    $shrunk =~ s/\A(\s*)//;
    $shrunk =~ s/(\s*)\Z//;
    if ($shrunk) {
	if ($self->{MODE} == $MODE_ByPredicate) {
	    if ($self->checkStackState(-1)) {
		my $predicate = $self->{STATE_STACK}[-1]->getPredicate;
		my $subject = $self->{STATE_STACK}[-1]->getNode;
		my $object = $self->{-rdfDB}->getString($shrunk);

		#			reification	reificationBag
		# 0	D1		I-1		B-2
		my ($reification, $container) = ($self->ensureReification(-1), $self->ensureReificationBag(-1));
		my $statement = $self->{-rdfDB}->addStatement ($predicate, $subject, $object, 
									      $container, $reification, $self->{ATTRIBUTION});
	    }
	} elsif ($self->{MODE} == $MODE_VALUE) {
	    if ($self->checkStackState(-2)) {
		my $predicate = $self->{STATE_STACK}[-1]->getPredicate;
		my $subject = $self->{STATE_STACK}[-2]->getNode;
		my $object = $self->{-rdfDB}->getString($shrunk);

		#			reification	reificationBag
		# 0	D1		I-1		B-2
		my ($reification, $container) = ($self->ensureReification(-1), $self->ensureReificationBag(-2));
		my $statement = $self->{-rdfDB}->addStatement ($predicate, $subject, $object, 
									      $container, $reification, $self->{ATTRIBUTION});
	    }
	}
    } elsif ($self->{MODE} == $MODE_LITTERAL) {
	$self->{CURRENT_LITTERAL} .= $ch;
    }
    $self->SUPER::characters($ch, $start, $length);
}

sub ignorableWhitespace {
    my ($self, $ch, $start, $length) = @_;
    $self->{STATE_STACK}[-1]->addSumString($ch) if ($self->{STATE_STACK}[-1]->getTag ne $self->{TRUST_SCHEMA_URI}.'sum');
}

sub show {
    my ($self, $prefix) = @_; $prefix = '  ' if (!defined $prefix);
    my $ret;
    $ret .= $prefix.$self."\n";
    $ret .= $prefix.'CONTAINER_ID: '.$self->{CONTAINER_ID}."\n";
#    $ret .= $prefix.'-trustedIds: '.$self->{-trustedIds}->dump."\n";
#    $ret .= $prefix.'-untrustedIds: '.$self->{-untrustedIds}->dump."\n";
    $ret .= $prefix."-trustedIds:\n";map {$ret .= $prefix.'  '.$_.': '.$self->{-trustedIds}{$_}->show."\n"} keys %{$self->{-trustedIds}};
    $ret .= $prefix."DEFAULT_DOC_TRUST:\n";map {$ret .= $prefix.'  '.$_.': '.$self->{DEFAULT_DOC_TRUST}{$_}."\n"} keys %{$self->{DEFAULT_DOC_TRUST}};
    $ret .= $prefix."DEFAULT_LINK_TRUST:\n";map {$ret .= $prefix.'  '.$_.': '.$self->{DEFAULT_LINK_TRUST}{$_}."\n"} keys %{$self->{DEFAULT_LINK_TRUST}};
    eval {$ret .= $self->{-rdfDB}->show($prefix.'  ')."\n";}; # don't assume the function exists
    return $ret;
}

1;

__END__

=head1 NAME

W3C::Rdf::RdfParser - parse RDF/XML and fill an RdfDB

=head1 SYNOPSIS

  use W3C::Rdf::RdfParser;
  use W3C::SAX::XmlParser;
  use W3C::SAX::InputSource;
  use W3C::SAX::XmlElement;
  use W3C::SAX::HandlerBase;
  
  &main(\@ARGV);
  
  sub main {
      my ($args) = @_;
      my ($source, @lookFors) = @$args; # ($args->[0], @$args[1..$#$args]);
  
      my $t0 = time; print "start\n";
  
      my $xmlParser = new W3C::SAX::PerlXmlParser;
  #    my $xmlParser = new W3C::SAX::ExpatXmlParser;
      my $errorHandler = new SAXErrorHandler($xmlParser);
      my $rdfDB = new W3C::Rdf::RdfDB($errorHandler, 1); # easier to read with lazy reification
      my $rdfParser = new W3C::Rdf::RdfParser(1, {}, {}, $rdfDB, undef, $errorHandler);
  
      # it's easiest to stick all your document handlers in a W3C::SAX::HandlerList
  #    my $handlerList = new W3C::SAX::HandlerList([new W3C::SAX::XmlElementTree, $rdfParser], [$errorHandler]);
      my $handlerList = new W3C::SAX::HandlerList([$rdfParser], [$errorHandler]);
  #    my $handlerList = new W3C::SAX::HandlerList([], [$errorHandler]);
  
      # use a W3C::SAX::NamespaceHandler stream to map from prefix:tag - pass to rdfParser
      # so it translate back for pretty-printing
      my $namespaceHandler = new W3C::SAX::NamespaceHandler($handlerList);
      $rdfParser->setNamespaceHandler($namespaceHandler);
  
      # and have the W3C::SAX::HandlerList be the document handler and the error handler
      $xmlParser->setDocumentHandler($namespaceHandler);
      $xmlParser->setErrorHandler($handlerList);
  
      # assign a URI to the XML in InputSorce and parse it
      my $inputSource = new W3C::SAX::FileInputSource($source);
      eval {$xmlParser->parse($inputSource);};
      if ($@) {
      	die $@.' at line '.$xmlParser->getLineNumber.' column '.$xmlParser->getColumnNumber;
      }
      my $t2 = time; print $t2-$t0,"\n";
  
      # some ways to show the result
  #    print $xmlParser->show('');
      print $rdfDB->showTriples('  ', $rdfParser); # give $rdfParser as a W3C::SAX::NamespaceHandler to unmap names.
      print "--------end of triples--------\n";
  
      # exercise some query functions in the W3C::Rdf::RdfDB
      @lookFors = ('http://www.w3.org/schema/certHTMLv1/access', 
  		 'http://www.w3.org/schema/certHTMLv1/accessor', 
  		 'http://www.w3.org/schema/certHTMLv1/hasAccessTo') if (!@lookFors);
      my @nodes = $rdfDB->nodesWithType(undef, 'http://www.w3.org/schema/certHTMLv1/resourceAccessRule');
  #    my @nodes = $rdfDB->nodesWithAllPredicates(undef, \@lookFors);
  
      # find all defined recombinations of the objects of the nodes
      my @join = $rdfDB->join(undef, \@nodes, \@lookFors);
  
      # show the onces we got
      foreach my $element (@join) {
  	for (my $i = $0; $i <= $#lookFors; $i++) {
  	    print $lookFors[$i].': '.$element->[$i].'  ';
  	}
  	print "\n";
      }
  
      # demonstrate the anySubjects function
      my @supplied = $rdfDB->supplied;
      my @theirStatements = $rdfDB->anySubjects(\@supplied, [$inputSource->getPublicId]);
      foreach my $triple (@theirStatements) {
  	print $triple->show."\n";
      }
  
      # test the anyContainerNodes function to make sure the acl stuff keeps working
      foreach my $node (@nodes) {
  	my @resourceList = $rdfDB->triplesMatching(undef, [['http://www.w3.org/schema/certHTMLv1/hasAccessTo', $node, undef]]);
  	my @resources;
  	map {push (@resources, $_->object)} @resourceList;
  	my @t = $rdfDB->anyContainerNodes(undef, [$node]);
  #	my @u = $rdfDB->hasAboutEach($node);
  	map {print "\$aclDB->addResourceAttribute(['".join('\',\'', @resources)."'], {'aclBinBag', ".$_->subject."})\n"} @t;
      }
  
      return;
  }
  
  package SAXErrorHandler;
  
  # this error handler is general enough to handle SAXExceptions and RdfDBExceptions
  
  sub new {
      my $class = shift;
      my $proto = ref $class || $class;
      my $self = {};
      bless($self, $proto);
      $self->{XML_PARSER} = shift;
      return $self;
  }
  
  sub makeString {
      my ($self, $exception) = @_;
      my $ret = $exception->toString;
      if ($exception->isa('W3C::Rdf::RdfDBException')) {
  	$ret .= ' at line '.$self->{XML_PARSER}->getLineNumber.' column '.$self->{XML_PARSER}->getColumnNumber;
      }
      return $ret;
  }
  
  sub warning {
      my ($self, $exception) = @_;
      print 'warning: '.$self->makeString($exception)."\n";
  }
  
  sub error {
      my ($self, $exception) = @_;
      warn 'error: '.$self->makeString($exception)."\n";
  }
  
  sub fatalError {
      my ($self, $exception) = @_;
      die 'fatalError: '.$self->makeString($exception)."\n";
  }

=head1 DESCRIPTION

W3C::SAX::XmlHandler to store RDF information in an RdfDB for later manipulation.

This module is part of the W3C::Rdf CPAN module.

=head1 AUTHOR

Eric Prud\'hommeaux <eric@w3.org>

=head1 SEE ALSO

W3C::Rdf::RdfDB(3) W3C::SAX::XmlParser(3) perl(1).

=cut
